# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import Atom
from portage.exception import InvalidAtom


class TestAtom(TestCase):
    def testAtom(self):
        tests = (
            (
                "=sys-apps/portage-2.1-r1:0[doc,a=,!b=,c?,!d?,-e]",
                (
                    "=",
                    "sys-apps/portage",
                    "2.1-r1",
                    "0",
                    "[doc,a=,!b=,c?,!d?,-e]",
                    None,
                ),
                False,
                False,
            ),
            (
                "=sys-apps/portage-2.1-r1*:0[doc]",
                ("=*", "sys-apps/portage", "2.1-r1", "0", "[doc]", None),
                False,
                False,
            ),
            (
                "sys-apps/portage:0[doc]",
                (None, "sys-apps/portage", None, "0", "[doc]", None),
                False,
                False,
            ),
            (
                "sys-apps/portage:0[doc]",
                (None, "sys-apps/portage", None, "0", "[doc]", None),
                False,
                False,
            ),
            ("*/*", (None, "*/*", None, None, None, None), True, False),
            ("=*/*-*9999*", ("=*", "*/*", "*9999*", None, None, None), True, False),
            (
                "=*/*-*9999*:0::repo_name",
                ("=*", "*/*", "*9999*", "0", None, "repo_name"),
                True,
                True,
            ),
            ("=*/*-*_beta*", ("=*", "*/*", "*_beta*", None, None, None), True, False),
            (
                "=*/*-*_beta*:0::repo_name",
                ("=*", "*/*", "*_beta*", "0", None, "repo_name"),
                True,
                True,
            ),
            ("sys-apps/*", (None, "sys-apps/*", None, None, None, None), True, False),
            ("*/portage", (None, "*/portage", None, None, None, None), True, False),
            (
                "s*s-*/portage:1",
                (None, "s*s-*/portage", None, "1", None, None),
                True,
                False,
            ),
            ("*/po*ge:2", (None, "*/po*ge", None, "2", None, None), True, False),
            ("!dev-libs/A", (None, "dev-libs/A", None, None, None, None), True, True),
            ("!!dev-libs/A", (None, "dev-libs/A", None, None, None, None), True, True),
            ("!!dev-libs/A", (None, "dev-libs/A", None, None, None, None), True, True),
            (
                "dev-libs/A[foo(+)]",
                (None, "dev-libs/A", None, None, "[foo(+)]", None),
                True,
                True,
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                (
                    None,
                    "dev-libs/A",
                    None,
                    None,
                    "[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                    None,
                ),
                True,
                True,
            ),
            (
                "dev-libs/A:2[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                (
                    None,
                    "dev-libs/A",
                    None,
                    "2",
                    "[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                    None,
                ),
                True,
                True,
            ),
            (
                "=sys-apps/portage-2.1-r1:0::repo_name[doc,a=,!b=,c?,!d?,-e]",
                (
                    "=",
                    "sys-apps/portage",
                    "2.1-r1",
                    "0",
                    "[doc,a=,!b=,c?,!d?,-e]",
                    "repo_name",
                ),
                False,
                True,
            ),
            (
                "=sys-apps/portage-2.1-r1*:0::repo_name[doc]",
                ("=*", "sys-apps/portage", "2.1-r1", "0", "[doc]", "repo_name"),
                False,
                True,
            ),
            (
                "sys-apps/portage:0::repo_name[doc]",
                (None, "sys-apps/portage", None, "0", "[doc]", "repo_name"),
                False,
                True,
            ),
            (
                "*/*::repo_name",
                (None, "*/*", None, None, None, "repo_name"),
                True,
                True,
            ),
            (
                "sys-apps/*::repo_name",
                (None, "sys-apps/*", None, None, None, "repo_name"),
                True,
                True,
            ),
            (
                "*/portage::repo_name",
                (None, "*/portage", None, None, None, "repo_name"),
                True,
                True,
            ),
            (
                "s*s-*/portage:1::repo_name",
                (None, "s*s-*/portage", None, "1", None, "repo_name"),
                True,
                True,
            ),
        )

        tests_xfail = (
            (Atom("sys-apps/portage"), False, False),
            ("cat/pkg[a!]", False, False),
            ("cat/pkg[!a]", False, False),
            ("cat/pkg[!a!]", False, False),
            ("cat/pkg[!a-]", False, False),
            ("cat/pkg[-a=]", False, False),
            ("cat/pkg[-a?]", False, False),
            ("cat/pkg[-a!]", False, False),
            ("cat/pkg[=a]", False, False),
            ("cat/pkg[=a=]", False, False),
            ("cat/pkg[=a?]", False, False),
            ("cat/pkg[=a!]", False, False),
            ("cat/pkg[=a-]", False, False),
            ("cat/pkg[?a]", False, False),
            ("cat/pkg[?a=]", False, False),
            ("cat/pkg[?a?]", False, False),
            ("cat/pkg[?a!]", False, False),
            ("cat/pkg[?a-]", False, False),
            ("sys-apps/portage[doc]:0", False, False),
            ("*/*", False, False),
            ("sys-apps/*", False, False),
            ("*/portage", False, False),
            ("*/**", True, False),
            ("*/portage[use]", True, False),
            ("cat/pkg[a()]", False, False),
            ("cat/pkg[a(]", False, False),
            ("cat/pkg[a)]", False, False),
            ("cat/pkg[a(,b]", False, False),
            ("cat/pkg[a),b]", False, False),
            ("cat/pkg[a(*)]", False, False),
            ("cat/pkg[a(*)]", True, False),
            ("cat/pkg[a(+-)]", False, False),
            ("cat/pkg[a()]", False, False),
            ("cat/pkg[(+)a]", False, False),
            ("cat/pkg[a=(+)]", False, False),
            ("cat/pkg[!(+)a=]", False, False),
            ("cat/pkg[!a=(+)]", False, False),
            ("cat/pkg[a?(+)]", False, False),
            ("cat/pkg[!a?(+)]", False, False),
            ("cat/pkg[!(+)a?]", False, False),
            ("cat/pkg[-(+)a]", False, False),
            ("cat/pkg[a(+),-a]", False, False),
            ("cat/pkg[a(-),-a]", False, False),
            ("cat/pkg[-a,a(+)]", False, False),
            ("cat/pkg[-a,a(-)]", False, False),
            ("cat/pkg[-a(+),a(-)]", False, False),
            ("cat/pkg[-a(-),a(+)]", False, False),
            ("sys-apps/portage[doc]::repo_name", False, False),
            ("sys-apps/portage:0[doc]::repo_name", False, False),
            ("sys-apps/portage[doc]:0::repo_name", False, False),
            (
                "=sys-apps/portage-2.1-r1:0::repo_name[doc,a=,!b=,c?,!d?,-e]",
                False,
                False,
            ),
            ("=sys-apps/portage-2.1-r1*:0::repo_name[doc]", False, False),
            ("sys-apps/portage:0::repo_name[doc]", False, False),
            ("*/*::repo_name", True, False),
        )

        for atom, parts, allow_wildcard, allow_repo in tests:
            a = Atom(atom, allow_wildcard=allow_wildcard, allow_repo=allow_repo)
            op, cp, ver, slot, use, repo = parts
            self.assertEqual(
                op,
                a.operator,
                msg=f"Atom('{atom}').operator = {a.operator} == '{op}'",
            )
            self.assertEqual(cp, a.cp, msg=f"Atom('{atom}').cp = {a.cp} == '{cp}'")
            if ver is not None:
                cpv = f"{cp}-{ver}"
            else:
                cpv = cp
            self.assertEqual(cpv, a.cpv, msg=f"Atom('{atom}').cpv = {a.cpv} == '{cpv}'")
            self.assertEqual(
                slot,
                a.slot,
                msg=f"Atom('{atom}').slot = {a.slot} == '{slot}'",
            )
            self.assertEqual(
                repo,
                a.repo,
                msg=f"Atom('{atom}').repo == {a.repo} == '{repo}'",
            )

            if a.use:
                returned_use = str(a.use)
            else:
                returned_use = None
            self.assertEqual(
                use,
                returned_use,
                msg=f"Atom('{atom}').use = {returned_use} == '{use}'",
            )

        for atom, allow_wildcard, allow_repo in tests_xfail:
            self.assertRaisesMsg(
                atom,
                (InvalidAtom, TypeError),
                Atom,
                atom,
                allow_wildcard=allow_wildcard,
                allow_repo=allow_repo,
            )

    def testSlotAbiAtom(self):
        tests = (
            (
                "virtual/ffmpeg:0/53",
                "5",
                {"slot": "0", "sub_slot": "53", "slot_operator": None},
            ),
            (
                "virtual/ffmpeg:0/53=",
                "5",
                {"slot": "0", "sub_slot": "53", "slot_operator": "="},
            ),
            (
                "virtual/ffmpeg:=",
                "5",
                {"slot": None, "sub_slot": None, "slot_operator": "="},
            ),
            (
                "virtual/ffmpeg:0=",
                "5",
                {"slot": "0", "sub_slot": None, "slot_operator": "="},
            ),
            (
                "virtual/ffmpeg:*",
                "5",
                {"slot": None, "sub_slot": None, "slot_operator": "*"},
            ),
            (
                "virtual/ffmpeg:0",
                "5",
                {"slot": "0", "sub_slot": None, "slot_operator": None},
            ),
            (
                "virtual/ffmpeg",
                "5",
                {"slot": None, "sub_slot": None, "slot_operator": None},
            ),
        )

        for atom, eapi, parts in tests:
            a = Atom(atom, eapi=eapi)
            for k, v in parts.items():
                self.assertEqual(
                    v,
                    getattr(a, k),
                    msg=f"Atom('{atom}').{k} = {getattr(a, k)} == '{v}'",
                )

    def test_intersects(self):
        test_cases = (
            ("dev-libs/A", "dev-libs/A", True),
            ("dev-libs/A", "dev-libs/B", False),
            ("dev-libs/A", "sci-libs/A", False),
            ("dev-libs/A[foo]", "sci-libs/A[bar]", False),
            ("dev-libs/A[foo(+)]", "sci-libs/A[foo(-)]", False),
            ("=dev-libs/A-1", "=dev-libs/A-1-r1", False),
            ("~dev-libs/A-1", "=dev-libs/A-1", False),
            ("=dev-libs/A-1:1", "=dev-libs/A-1", True),
            ("=dev-libs/A-1:1", "=dev-libs/A-1:1", True),
            ("=dev-libs/A-1:1", "=dev-libs/A-1:2", False),
        )

        for atom, other, expected_result in test_cases:
            self.assertEqual(
                Atom(atom).intersects(Atom(other)),
                expected_result,
                f"{atom} and {other} should intersect: {expected_result}",
            )

    def test_violated_conditionals(self):
        test_cases = (
            ("dev-libs/A", ["foo"], ["foo"], None, "dev-libs/A"),
            ("dev-libs/A[foo]", [], ["foo"], None, "dev-libs/A[foo]"),
            ("dev-libs/A[foo]", ["foo"], ["foo"], None, "dev-libs/A"),
            ("dev-libs/A[foo]", [], ["foo"], [], "dev-libs/A[foo]"),
            ("dev-libs/A[foo]", ["foo"], ["foo"], [], "dev-libs/A"),
            ("dev-libs/A:0[foo]", ["foo"], ["foo"], [], "dev-libs/A:0"),
            ("dev-libs/A[foo,-bar]", [], ["foo", "bar"], None, "dev-libs/A[foo]"),
            ("dev-libs/A[-foo,bar]", [], ["foo", "bar"], None, "dev-libs/A[bar]"),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                [],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["a"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["b"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a,b=,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["c"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["d"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["e"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a,!c=,!e?]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["f"],
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a,!c=,-f]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["a"],
                ["a", "b", "c", "d", "e", "f"],
                ["a"],
                "dev-libs/A[!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["b"],
                ["a", "b", "c", "d", "e", "f"],
                ["b"],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["c"],
                ["a", "b", "c", "d", "e", "f"],
                ["c"],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["d"],
                ["a", "b", "c", "d", "e", "f"],
                ["d"],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["e"],
                ["a", "b", "c", "d", "e", "f"],
                ["e"],
                "dev-libs/A[a,!c=]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["f"],
                ["a", "b", "c", "d", "e", "f"],
                ["f"],
                "dev-libs/A[a,!c=,-f]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["a"],
                ["a", "b", "c", "d", "e", "f"],
                ["a"],
                "dev-libs/A[!c(+)=]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["b"],
                ["a", "b", "c", "d", "e", "f"],
                ["b"],
                "dev-libs/A[a(-),!c(-)=]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["c"],
                ["a", "b", "c", "d", "e", "f"],
                ["c"],
                "dev-libs/A[a(+),!c(+)=]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["d"],
                ["a", "b", "c", "d", "e", "f"],
                ["d"],
                "dev-libs/A[a(-),!c(-)=]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["e"],
                ["a", "b", "c", "d", "e", "f"],
                ["e"],
                "dev-libs/A[a(+),!c(+)=]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["f"],
                ["a", "b", "c", "d", "e", "f"],
                ["f"],
                "dev-libs/A[a(-),!c(-)=,-f(+)]",
            ),
            (
                "dev-libs/A[a(+),b(+)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["a"],
                ["a"],
                ["a"],
                "dev-libs/A[b(+)=,!e(+)?]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["b"],
                ["b"],
                ["b"],
                "dev-libs/A[a(-),!c(-)=,-f(+)]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["c"],
                ["c"],
                ["c"],
                "dev-libs/A[!c(+)=,!e(+)?]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["d"],
                ["d"],
                ["d"],
                "dev-libs/A[a(-),b(+)=,!c(-)=,-f(+)]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["e"],
                ["e"],
                ["e"],
                "dev-libs/A",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["f"],
                ["f"],
                ["f"],
                "dev-libs/A[a(-),b(+)=,!c(-)=,-f(+)]",
            ),
            # Some more test cases to trigger all remaining code paths
            ("dev-libs/B[x?]", [], ["x"], ["x"], "dev-libs/B[x?]"),
            ("dev-libs/B[x(+)?]", [], [], ["x"], "dev-libs/B"),
            ("dev-libs/B[x(-)?]", [], [], ["x"], "dev-libs/B[x(-)?]"),
            ("dev-libs/C[x=]", [], ["x"], ["x"], "dev-libs/C[x=]"),
            ("dev-libs/C[x(+)=]", [], [], ["x"], "dev-libs/C"),
            ("dev-libs/C[x(-)=]", [], [], ["x"], "dev-libs/C[x(-)=]"),
            ("dev-libs/D[!x=]", [], ["x"], ["x"], "dev-libs/D"),
            ("dev-libs/D[!x(+)=]", [], [], ["x"], "dev-libs/D[!x(+)=]"),
            ("dev-libs/D[!x(-)=]", [], [], ["x"], "dev-libs/D"),
            # Missing IUSE test cases
            ("dev-libs/B[x]", [], [], [], "dev-libs/B[x]"),
            ("dev-libs/B[-x]", [], [], [], "dev-libs/B[-x]"),
            ("dev-libs/B[x?]", [], [], [], "dev-libs/B[x?]"),
            ("dev-libs/B[x=]", [], [], [], "dev-libs/B[x=]"),
            ("dev-libs/B[!x=]", [], [], ["x"], "dev-libs/B[!x=]"),
            ("dev-libs/B[!x?]", [], [], ["x"], "dev-libs/B[!x?]"),
        )

        test_cases_xfail = (
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                [],
                ["a", "b", "c", "d", "e", "f"],
                None,
            ),
        )

        class use_flag_validator:
            def __init__(self, iuse):
                self.iuse = iuse

            def is_valid_flag(self, flag):
                return flag in iuse

        for atom, other_use, iuse, parent_use, expected_violated_atom in test_cases:
            a = Atom(atom)
            validator = use_flag_validator(iuse)
            violated_atom = a.violated_conditionals(
                other_use, validator.is_valid_flag, parent_use
            )
            if parent_use is None:
                fail_msg = (
                    "Atom: %s, other_use: %s, iuse: %s, parent_use: %s, got: %s, expected: %s"
                    % (
                        atom,
                        " ".join(other_use),
                        " ".join(iuse),
                        "None",
                        str(violated_atom),
                        expected_violated_atom,
                    )
                )
            else:
                fail_msg = (
                    "Atom: %s, other_use: %s, iuse: %s, parent_use: %s, got: %s, expected: %s"
                    % (
                        atom,
                        " ".join(other_use),
                        " ".join(iuse),
                        " ".join(parent_use),
                        str(violated_atom),
                        expected_violated_atom,
                    )
                )
            self.assertEqual(str(violated_atom), expected_violated_atom, fail_msg)

        for atom, other_use, iuse, parent_use in test_cases_xfail:
            a = Atom(atom)
            validator = use_flag_validator(iuse)
            self.assertRaisesMsg(
                atom,
                InvalidAtom,
                a.violated_conditionals,
                other_use,
                validator.is_valid_flag,
                parent_use,
            )

    def test_evaluate_conditionals(self):
        test_cases = (
            ("dev-libs/A[foo]", [], "dev-libs/A[foo]"),
            ("dev-libs/A[foo]", ["foo"], "dev-libs/A[foo]"),
            ("dev-libs/A:0[foo=]", ["foo"], "dev-libs/A:0[foo]"),
            ("dev-libs/A[foo,-bar]", [], "dev-libs/A[foo,-bar]"),
            ("dev-libs/A[-foo,bar]", [], "dev-libs/A[-foo,bar]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", [], "dev-libs/A[a,-b,c,-e,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["a"], "dev-libs/A[a,-b,c,-e,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["b"], "dev-libs/A[a,b,c,-e,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["c"], "dev-libs/A[a,-b,-c,-e,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["d"], "dev-libs/A[a,-b,c,d,-e,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["e"], "dev-libs/A[a,-b,c,-f]"),
            ("dev-libs/A[a,b=,!c=,d?,!e?,-f]", ["f"], "dev-libs/A[a,-b,c,-e,-f]"),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["d"],
                "dev-libs/A[a(-),-b(+),c(-),d(+),-e(-),-f(+)]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                ["f"],
                "dev-libs/A[a(+),-b(-),c(+),-e(+),-f(-)]",
            ),
        )

        for atom, use, expected_atom in test_cases:
            a = Atom(atom)
            b = a.evaluate_conditionals(use)
            self.assertEqual(str(b), expected_atom)
            self.assertEqual(str(b.unevaluated_atom), atom)

    def test__eval_qa_conditionals(self):
        test_cases = (
            ("dev-libs/A[foo]", [], [], "dev-libs/A[foo]"),
            ("dev-libs/A[foo]", ["foo"], [], "dev-libs/A[foo]"),
            ("dev-libs/A[foo]", [], ["foo"], "dev-libs/A[foo]"),
            ("dev-libs/A:0[foo]", [], [], "dev-libs/A:0[foo]"),
            ("dev-libs/A:0[foo]", ["foo"], [], "dev-libs/A:0[foo]"),
            ("dev-libs/A:0[foo]", [], ["foo"], "dev-libs/A:0[foo]"),
            ("dev-libs/A:0[foo=]", [], ["foo"], "dev-libs/A:0[foo]"),
            ("dev-libs/A[foo,-bar]", ["foo"], ["bar"], "dev-libs/A[foo,-bar]"),
            ("dev-libs/A[-foo,bar]", ["foo", "bar"], [], "dev-libs/A[-foo,bar]"),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["a", "b", "c"],
                [],
                "dev-libs/A[a,-b,c,d,-e,-f]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                [],
                ["a", "b", "c"],
                "dev-libs/A[a,b,-c,d,-e,-f]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                ["d", "e", "f"],
                [],
                "dev-libs/A[a,b,-b,c,-c,-e,-f]",
            ),
            (
                "dev-libs/A[a,b=,!c=,d?,!e?,-f]",
                [],
                ["d", "e", "f"],
                "dev-libs/A[a,b,-b,c,-c,d,-f]",
            ),
            (
                "dev-libs/A[a(-),b(+)=,!c(-)=,d(+)?,!e(-)?,-f(+)]",
                ["a", "b", "c", "d", "e", "f"],
                [],
                "dev-libs/A[a(-),-b(+),c(-),-e(-),-f(+)]",
            ),
            (
                "dev-libs/A[a(+),b(-)=,!c(+)=,d(-)?,!e(+)?,-f(-)]",
                [],
                ["a", "b", "c", "d", "e", "f"],
                "dev-libs/A[a(+),b(-),-c(+),d(-),-f(-)]",
            ),
        )

        for atom, use_mask, use_force, expected_atom in test_cases:
            a = Atom(atom)
            b = a._eval_qa_conditionals(use_mask, use_force)
            self.assertEqual(str(b), expected_atom)
            self.assertEqual(str(b.unevaluated_atom), atom)



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OrUpgradeInstalledTestCase(TestCase):
    def testOrUpgradeInstalled(self):
        ebuilds = {
            "net-misc/foo-1": {
                "EAPI": "6",
                "RDEPEND": "|| ( sys-libs/glibc[rpc(-)]  net-libs/libtirpc )",
            },
            "net-libs/libtirpc-1": {
                "EAPI": "6",
            },
            "sys-libs/glibc-2.26": {"EAPI": "6", "IUSE": ""},
            "sys-libs/glibc-2.24": {"EAPI": "6", "IUSE": "+rpc"},
        }

        installed = {
            "sys-libs/glibc-2.24": {
                "EAPI": "6",
                "IUSE": "+rpc",
                "USE": "rpc",
            },
        }

        world = ["sys-libs/glibc"]

        test_cases = (
            # Test bug 643974, where we need to install libtirpc
            # in order to upgrade glibc.
            ResolverPlaygroundTestCase(
                ["net-misc/foo", "@world"],
                options={"--update": True, "--deep": True},
                success=True,
                ambiguous_merge_order=True,
                mergelist=(
                    (
                        "net-libs/libtirpc-1",
                        "sys-libs/glibc-2.26",
                        "net-misc/foo-1",
                    ),
                ),
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        # In some cases it's necessary to avoid upgrade due to
        # the package being masked.
        user_config = {
            "package.mask": (">=sys-libs/glibc-2.26",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["net-misc/foo", "@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "net-misc/foo-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False,
            ebuilds=ebuilds,
            installed=installed,
            world=world,
            user_config=user_config,
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testVirtualRust(self):
        ebuilds = {
            "dev-lang/rust-1.19.0": {},
            "dev-lang/rust-1.23.0": {},
            "dev-lang/rust-bin-1.19.0": {},
            "virtual/rust-1.19.0": {
                "RDEPEND": "|| ( =dev-lang/rust-1.19.0* =dev-lang/rust-bin-1.19.0* )"
            },
        }

        installed = {
            "dev-lang/rust-1.19.0": {},
            "virtual/rust-1.19.0": {
                "RDEPEND": "|| ( =dev-lang/rust-1.19.0* =dev-lang/rust-bin-1.19.0* )"
            },
        }

        world = ["virtual/rust"]

        test_cases = (
            # Test bug 645416, where rust-bin-1.19.0 was pulled in
            # inappropriately due to the rust-1.23.0 update being
            # available.
            ResolverPlaygroundTestCase(
                ["virtual/rust"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
            # Test upgrade to rust-1.23.0, which is only possible
            # if rust-bin-1.19.0 is installed in order to satisfy
            # virtual/rust-1.19.0.
            ResolverPlaygroundTestCase(
                ["=dev-lang/rust-1.23.0", "virtual/rust"],
                options={"--update": True, "--deep": True},
                all_permutations=True,
                success=True,
                ambiguous_merge_order=True,
                mergelist=(
                    (
                        "dev-lang/rust-1.23.0",
                        "dev-lang/rust-bin-1.19.0",
                    ),
                ),
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def test_llvm_slot_operator(self):
        ebuilds = {
            "media-libs/mesa-19.2.8": {
                "EAPI": "7",
                "RDEPEND": """|| (
					sys-devel/llvm:10
					sys-devel/llvm:9
					sys-devel/llvm:8
					sys-devel/llvm:7
				)
				sys-devel/llvm:=""",
            },
            "sys-devel/llvm-10": {
                "EAPI": "7",
                "KEYWORDS": "",
                "SLOT": "10",
            },
            "sys-devel/llvm-9": {
                "EAPI": "7",
                "SLOT": "9",
            },
            "sys-devel/llvm-8": {
                "EAPI": "7",
                "SLOT": "8",
            },
        }

        installed = {
            "media-libs/mesa-19.2.8": {
                "EAPI": "7",
                "RDEPEND": """|| (
					sys-devel/llvm:10
					sys-devel/llvm:9
					sys-devel/llvm:8
					sys-devel/llvm:7
				)
				sys-devel/llvm:8/8=""",
            },
            "sys-devel/llvm-8": {
                "EAPI": "7",
                "SLOT": "8",
            },
        }

        world = ["media-libs/mesa"]

        test_cases = (
            # Demonstrate bug 706278, where there is a missed slot operator
            # rebuild that prevents upgrade from llvm-8 to llvm-9.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["sys-devel/llvm-9", "media-libs/mesa-19.2.8"],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2020-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import os
import shutil
import tempfile
import time

from portage.tests import TestCase
from portage.util.shelve import dump, open_shelve, restore


class ShelveUtilsTestCase(TestCase):
    TEST_DATA = (
        # distfiles_db
        {
            "portage-2.3.89.tar.bz2": "sys-apps/portage-2.3.89",
            "portage-2.3.99.tar.bz2": "sys-apps/portage-2.3.99",
        },
        # deletion_db
        {
            "portage-2.3.89.tar.bz2": time.time(),
            "portage-2.3.99.tar.bz2": time.time(),
        },
        # recycle_db
        {
            "portage-2.3.89.tar.bz2": (0, time.time()),
            "portage-2.3.99.tar.bz2": (0, time.time()),
        },
    )

    def test_dump_restore(self):
        for data in self.TEST_DATA:
            tmpdir = tempfile.mkdtemp()
            try:
                dump_args = argparse.Namespace(
                    src=os.path.join(tmpdir, "shelve_file"),
                    dest=os.path.join(tmpdir, "pickle_file"),
                )
                db = open_shelve(dump_args.src, flag="c")
                for k, v in data.items():
                    db[k] = v
                db.close()
                dump(dump_args)

                restore_args = argparse.Namespace(
                    dest=dump_args.src,
                    src=dump_args.dest,
                )
                restore(restore_args)

                db = open_shelve(restore_args.dest, flag="r")
                for k, v in data.items():
                    self.assertEqual(db[k], v)
                db.close()
            finally:
                shutil.rmtree(tmpdir)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
from portage.util import writemsg
from portage.data import secpass
import portage
from portage import os
import pickle


class BlockerCache(portage.cache.mappings.MutableMapping):
    """This caches blockers of installed packages so that dep_check does not
    have to be done for every single installed package on every invocation of
    emerge.  The cache is invalidated whenever it is detected that something
    has changed that might alter the results of dep_check() calls:
            1) the set of installed packages (including COUNTER) has changed
    """

    # Number of uncached packages to trigger cache update, since
    # it's wasteful to update it for every vdb change.
    _cache_threshold = 5

    class BlockerData:
        __slots__ = ("__weakref__", "atoms", "counter")

        def __init__(self, counter, atoms):
            self.counter = counter
            self.atoms = atoms

    def __init__(self, myroot, vardb):
        """myroot is ignored in favour of EROOT"""
        self._vardb = vardb
        self._cache_filename = os.path.join(
            vardb.settings["EROOT"], portage.CACHE_PATH, "vdb_blockers.pickle"
        )
        self._cache_version = "1"
        self._cache_data = None
        self._modified = set()
        self._load()

    def _load(self):
        try:
            f = open(self._cache_filename, mode="rb")
            mypickle = pickle.Unpickler(f)
            try:
                mypickle.find_global = None
            except AttributeError:
                # TODO: If py3k, override Unpickler.find_class().
                pass
            self._cache_data = mypickle.load()
            f.close()
            del f
        except (SystemExit, KeyboardInterrupt):
            raise
        except Exception as e:
            if isinstance(e, EnvironmentError) and getattr(e, "errno", None) in (
                errno.ENOENT,
                errno.EACCES,
            ):
                pass
            else:
                writemsg(
                    f"!!! Error loading '{self._cache_filename}': {str(e)}\n",
                    noiselevel=-1,
                )
            del e

        cache_valid = (
            self._cache_data
            and isinstance(self._cache_data, dict)
            and self._cache_data.get("version") == self._cache_version
            and isinstance(self._cache_data.get("blockers"), dict)
        )
        if cache_valid:
            # Validate all the atoms and counters so that
            # corruption is detected as soon as possible.
            invalid_items = set()
            for k, v in self._cache_data["blockers"].items():
                if not isinstance(k, str):
                    invalid_items.add(k)
                    continue
                try:
                    if portage.catpkgsplit(k) is None:
                        invalid_items.add(k)
                        continue
                except portage.exception.InvalidData:
                    invalid_items.add(k)
                    continue
                if not isinstance(v, tuple) or len(v) != 2:
                    invalid_items.add(k)
                    continue
                counter, atoms = v
                if not isinstance(counter, int):
                    invalid_items.add(k)
                    continue
                if not isinstance(atoms, (list, tuple)):
                    invalid_items.add(k)
                    continue
                invalid_atom = False
                for atom in atoms:
                    if not isinstance(atom, str):
                        invalid_atom = True
                        break
                    if atom[:1] != "!" or not portage.isvalidatom(
                        atom, allow_blockers=True
                    ):
                        invalid_atom = True
                        break
                if invalid_atom:
                    invalid_items.add(k)
                    continue

            for k in invalid_items:
                del self._cache_data["blockers"][k]
            if not self._cache_data["blockers"]:
                cache_valid = False

        if not cache_valid:
            self._cache_data = {"version": self._cache_version}
            self._cache_data["blockers"] = {}
        self._modified.clear()

    def flush(self):
        """If the current user has permission and the internal blocker cache has
        been updated, save it to disk and mark it unmodified.  This is called
        by emerge after it has processed blockers for all installed packages.
        Currently, the cache is only written if the user has superuser
        privileges (since that's required to obtain a lock), but all users
        have read access and benefit from faster blocker lookups (as long as
        the entire cache is still valid).  The cache is stored as a pickled
        dict object with the following format:

        {
                version : "1",
                "blockers" : {cpv1:(counter,(atom1, atom2...)), cpv2...},
        }
        """
        if len(self._modified) >= self._cache_threshold and secpass >= 2:
            try:
                with portage.util.atomic_ofstream(self._cache_filename, mode="wb") as f:
                    pickle.dump(self._cache_data, f, protocol=2)

                portage.util.apply_secpass_permissions(
                    self._cache_filename, gid=portage.portage_gid, mode=0o644
                )
            except OSError:
                pass
            self._modified.clear()

    def __setitem__(self, cpv, blocker_data):
        """
        Update the cache and mark it as modified for a future call to
        self.flush().

        @param cpv: Package for which to cache blockers.
        @type cpv: String
        @param blocker_data: An object with counter and atoms attributes.
        @type blocker_data: BlockerData
        """
        self._cache_data["blockers"][str(cpv)] = (
            blocker_data.counter,
            tuple(str(x) for x in blocker_data.atoms),
        )
        self._modified.add(cpv)

    def __iter__(self):
        if self._cache_data is None:
            # triggered by python-trace
            return iter([])
        return iter(self._cache_data["blockers"])

    def __len__(self):
        """This needs to be implemented in order to avoid
        infinite recursion in some cases."""
        return len(self._cache_data["blockers"])

    def __delitem__(self, cpv):
        del self._cache_data["blockers"][cpv]

    def __getitem__(self, cpv):
        """
        @rtype: BlockerData
        @return: An object with counter and atoms attributes.
        """
        return self.BlockerData(*self._cache_data["blockers"][cpv])



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskUseBacktrackTestCase(TestCase):
    def testAutounmaskUseBacktrack(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "6",
                "RDEPEND": "dev-libs/C",
            },
            "dev-libs/A-2": {
                "EAPI": "6",
                "RDEPEND": "dev-libs/C[y]",
            },
            "dev-libs/A-3": {
                "EAPI": "6",
                "RDEPEND": "dev-libs/C",
            },
            "dev-libs/B-1": {
                "EAPI": "6",
                "RDEPEND": "<dev-libs/A-3",
            },
            "dev-libs/C-1": {
                "EAPI": "6",
                "IUSE": "x y z",
            },
            "dev-libs/D-1": {
                "EAPI": "6",
                "RDEPEND": ">=dev-libs/A-2 dev-libs/C[x]",
            },
        }

        installed = {
            "dev-libs/A-1": {
                "EAPI": "6",
                "RDEPEND": "dev-libs/C",
            },
            "dev-libs/B-1": {
                "EAPI": "6",
                "RDEPEND": "<dev-libs/A-3",
            },
            "dev-libs/C-1": {
                "EAPI": "6",
                "IUSE": "x y z",
            },
        }

        world = ["dev-libs/B"]

        test_cases = (
            # Test bug 632598, where autounmask USE changes triggered
            # unnecessary backtracking. The following case should
            # require a --backtrack setting no larger than 2.
            ResolverPlaygroundTestCase(
                ["dev-libs/D"],
                options={
                    "--autounmask-backtrack": "y",
                    "--backtrack": 2,
                },
                success=False,
                ambiguous_merge_order=True,
                mergelist=[
                    ("dev-libs/C-1", "dev-libs/A-2"),
                    "dev-libs/D-1",
                ],
                use_changes={"dev-libs/C-1": {"y": True, "x": True}},
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from unittest.mock import MagicMock, patch, call
import os
import tempfile

from portage.tests import TestCase

from portage.dbapi.bintree import binarytree
from portage.localization import _
from portage.const import BINREPOS_CONF_FILE


class BinarytreeTestCase(TestCase):
    def test_required_init_params(self):
        with self.assertRaises(TypeError) as cm:
            binarytree()
        self.assertEqual(str(cm.exception), "pkgdir parameter is required")
        with self.assertRaises(TypeError) as cm:
            binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"))
        self.assertEqual(str(cm.exception), "settings parameter is required")

    def test_init_with_legacy_params_warns(self):
        with self.assertWarns(DeprecationWarning):
            binarytree(
                _unused=None, pkgdir=os.getenv("TMPDIR", "/tmp"), settings=MagicMock()
            )
        with self.assertWarns(DeprecationWarning):
            binarytree(
                virtual=None, pkgdir=os.getenv("TMPDIR", "/tmp"), settings=MagicMock()
            )

    def test_instance_has_required_attrs(self):
        # Quite smoky test. What would it be a better testing strategy?
        # Not sure yet...
        required_attrs_no_multi_instance = {
            "pkgdir",
            "_multi_instance",
            "dbapi",
            "update_ents",
            "move_slot_ent",
            "populated",
            "tree",
            "_binrepos_conf",
            "_remote_has_index",
            "_remotepkgs",
            "_additional_pkgs",
            "invalids",
            "invalid_paths",
            "settings",
            "_pkg_paths",
            "_populating",
            "_all_directory",
            "_pkgindex_version",
            "_pkgindex_hashes",
            "_pkgindex_file",
            "_pkgindex_keys",
            "_pkgindex_aux_keys",
            "_pkgindex_use_evaluated_keys",
            "_pkgindex_header",
            "_pkgindex_header_keys",
            "_pkgindex_default_pkg_data",
            "_pkgindex_inherited_keys",
            "_pkgindex_default_header_data",
            "_pkgindex_translated_keys",
            "_pkgindex_allowed_pkg_keys",
        }
        no_multi_instance_settings = MagicMock()
        no_multi_instance_settings.features = ""
        no_multi_instance_bt = binarytree(
            pkgdir=os.getenv("TMPDIR", "/tmp"), settings=no_multi_instance_settings
        )
        multi_instance_settings = MagicMock()
        multi_instance_settings.features = "binpkg-multi-instance"
        multi_instance_bt = binarytree(
            pkgdir=os.getenv("TMPDIR", "/tmp"), settings=multi_instance_settings
        )
        for attr in required_attrs_no_multi_instance:
            getattr(no_multi_instance_bt, attr)
            getattr(multi_instance_bt, attr)
        # The next attribute is the difference between multi instance
        # and no multi instance:
        getattr(multi_instance_bt, "_allocate_filename")

    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_populate_without_updates_repos_nor_getbinspkgs(self, ppopulate_local):
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=MagicMock())
        ppopulate_local.return_value = {}
        bt.populate()
        ppopulate_local.assert_called_once_with(reindex=True, invalid_errors=True)
        self.assertFalse(bt._populating)
        self.assertTrue(bt.populated)

    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_populate_calls_twice_populate_local_if_updates(self, ppopulate_local):
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=MagicMock())
        bt.populate()
        self.assertIn(
            call(reindex=True, invalid_errors=True), ppopulate_local.mock_calls
        )
        self.assertIn(call(), ppopulate_local.mock_calls)
        self.assertEqual(ppopulate_local.call_count, 2)

    @patch("portage.dbapi.bintree.binarytree._populate_additional")
    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_populate_with_repos(self, ppopulate_local, ppopulate_additional):
        repos = ("one", "two")
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=MagicMock())
        bt.populate(add_repos=repos)
        ppopulate_additional.assert_called_once_with(repos)

    @patch("portage.dbapi.bintree.BinRepoConfigLoader")
    @patch("portage.dbapi.bintree.binarytree._populate_remote")
    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_populate_with_getbinpkgs(
        self, ppopulate_local, ppopulate_remote, pBinRepoConfigLoader
    ):
        refresh = "something"
        settings = MagicMock()
        settings.__getitem__.return_value = "/some/path"
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=settings)
        bt.populate(getbinpkgs=True, getbinpkg_refresh=refresh)
        ppopulate_remote.assert_called_once_with(
            getbinpkg_refresh=refresh, pretend=False
        )

    @patch("portage.dbapi.bintree.writemsg")
    @patch("portage.dbapi.bintree.BinRepoConfigLoader")
    @patch("portage.dbapi.bintree.binarytree._populate_remote")
    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_populate_with_getbinpkgs_and_not_BinRepoConfigLoader(
        self, ppopulate_local, ppopulate_remote, pBinRepoConfigLoader, pwritemsg
    ):
        refresh = "something"
        settings = MagicMock()
        portage_root = "/some/path"
        settings.__getitem__.return_value = portage_root
        pBinRepoConfigLoader.return_value = None
        conf_file = os.path.join(portage_root, BINREPOS_CONF_FILE)
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=settings)
        bt.populate(getbinpkgs=True, getbinpkg_refresh=refresh)
        ppopulate_remote.assert_not_called()
        pwritemsg.assert_called_once_with(
            _(
                f"!!! {conf_file} is missing (or PORTAGE_BINHOST is unset)"
                ", but use is requested.\n"
            ),
            noiselevel=-1,
        )

    @patch("portage.dbapi.bintree.BinRepoConfigLoader")
    @patch("portage.dbapi.bintree.binarytree._populate_remote")
    @patch("portage.dbapi.bintree.binarytree._populate_local")
    def test_default_getbinpkg_refresh_in_populate(
        self, ppopulate_local, ppopulate_remote, pBinRepoConfigLoader
    ):
        """Bug #864259
        This test fixes the bug. It requires that
        ``_emerge.actions.run_action`` calls ``binarytree.populate``
        explicitly with ``getbinpkg_refresh=True``
        """
        settings = MagicMock()
        settings.__getitem__.return_value = "/some/path"
        bt = binarytree(pkgdir=os.getenv("TMPDIR", "/tmp"), settings=settings)
        bt.populate(getbinpkgs=True)
        ppopulate_remote.assert_called_once_with(getbinpkg_refresh=False, pretend=False)

    @patch("portage.dbapi.bintree.BinRepoConfigLoader")
    @patch("portage.dbapi.bintree.binarytree._run_trust_helper")
    def test_default_getbinpkg_refresh_in_populate_trusthelper(
        self, run_trust_helper, pBinRepoConfigLoader
    ):
        """
        Test for bug #915842.

        Verify that we call the trust helper in non-pretend mode.
        """
        settings = MagicMock()
        settings.features = ["binpkg-request-signature"]
        settings.__getitem__.return_value = "/some/path"

        d = tempfile.TemporaryDirectory()
        try:
            bt = binarytree(pkgdir=d.name, settings=settings)
            bt.populate(getbinpkgs=True, pretend=False)
            run_trust_helper.assert_called_once()
        finally:
            d.cleanup()

    @patch("portage.dbapi.bintree.BinRepoConfigLoader")
    @patch("portage.dbapi.bintree.binarytree._run_trust_helper")
    def test_default_getbinpkg_refresh_in_populate_trusthelper_pretend(
        self, run_trust_helper, pBinRepoConfigLoader
    ):
        """
        Test for bug #915842.

        Verify we do not call the trust helper in pretend mode.
        """
        settings = MagicMock()
        settings.features = ["binpkg-request-signature"]
        settings.__getitem__.return_value = "/some/path"

        d = tempfile.TemporaryDirectory()
        try:
            bt = binarytree(pkgdir=d.name, settings=settings)
            bt.populate(getbinpkgs=True, pretend=True)
            run_trust_helper.assert_not_called()
        finally:
            d.cleanup()



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from portage.process import find_binary, spawn
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures.unix_events import DefaultEventLoopPolicy


class ChildWatcherTestCase(TestCase):
    def testChildWatcher(self):
        true_binary = find_binary("true")
        self.assertNotEqual(true_binary, None)

        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        loop = None
        try:
            try:
                asyncio.set_child_watcher(None)
            except NotImplementedError:
                pass
            else:
                self.assertTrue(False)

            args_tuple = ("hello", "world")

            loop = asyncio._wrap_loop()
            future = loop.create_future()

            def callback(pid, returncode, *args):
                future.set_result((pid, returncode, args))

            async def watch_pid():
                with asyncio.get_child_watcher() as watcher:
                    pids = spawn([true_binary], returnpid=True)
                    watcher.add_child_handler(pids[0], callback, *args_tuple)
                    self.assertEqual((await future), (pids[0], os.EX_OK, args_tuple))

            loop.run_until_complete(watch_pid())
        finally:
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractEbuildProcess import AbstractEbuildProcess
import portage

portage.proxy.lazyimport.lazyimport(globals(), "portage.package.ebuild.doebuild:spawn")
from portage import os


class MiscFunctionsProcess(AbstractEbuildProcess):
    """
    Spawns misc-functions.sh with an existing ebuild environment.
    """

    __slots__ = ("commands", "ld_preload_sandbox")

    def _start(self):
        settings = self.settings
        portage_bin_path = settings["PORTAGE_BIN_PATH"]
        misc_sh_binary = os.path.join(
            portage_bin_path, os.path.basename(portage.const.MISC_SH_BINARY)
        )

        self.args = [portage._shell_quote(misc_sh_binary)] + self.commands
        if (
            self.logfile is None
            and self.settings.get("PORTAGE_BACKGROUND") != "subprocess"
        ):
            self.logfile = settings.get("PORTAGE_LOG_FILE")

        AbstractEbuildProcess._start(self)

    def _spawn(self, args, **kwargs):
        # If self.ld_preload_sandbox is None, default to free=False,
        # in alignment with the spawn(free=False) default.
        kwargs.setdefault(
            "free",
            False if self.ld_preload_sandbox is None else not self.ld_preload_sandbox,
        )

        if self._dummy_pipe_fd is not None:
            self.settings["PORTAGE_PIPE_FD"] = str(self._dummy_pipe_fd)

        if "fakeroot" in self.settings.features:
            kwargs["fakeroot"] = True

        # Temporarily unset EBUILD_PHASE so that bashrc code doesn't
        # think this is a real phase.
        phase_backup = self.settings.pop("EBUILD_PHASE", None)

        # Use a separate sandbox log file to avoid clobbering the real one.
        sandbox_log = self.settings.pop("SANDBOX_LOG", None)
        if sandbox_log is not None:
            self.settings["SANDBOX_LOG"] = os.path.join(
                self.settings["T"], "sandbox-misc.log"
            )

        try:
            return spawn(" ".join(args), self.settings, **kwargs)
        finally:
            if phase_backup is not None:
                self.settings["EBUILD_PHASE"] = phase_backup
            self.settings.pop("PORTAGE_PIPE_FD", None)
            self.settings.pop("SANDBOX_LOG", None)
            if sandbox_log is not None:
                self.settings["SANDBOX_LOG"] = sandbox_log



# Copyright 2010-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import fcntl
import logging
import sys
import threading

import portage
from portage import os
from portage.exception import TryAgain
from portage.locks import lockfile, unlockfile
from portage.util import writemsg_level
from _emerge.AbstractPollTask import AbstractPollTask
from _emerge.AsynchronousTask import AsynchronousTask
from _emerge.SpawnProcess import SpawnProcess


class AsynchronousLock(AsynchronousTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using either a thread (if available) or a subprocess.

    The default behavior is to use a process instead of a thread, since
    there is currently no way to interrupt a thread that is waiting for
    a lock (notably, SIGINT doesn't work because python delivers all
    signals to the main thread).
    """

    __slots__ = ("path",) + (
        "_imp",
        "_force_async",
        "_force_process",
        "_force_thread",
        "_unlock_future",
    )

    _use_process_by_default = True

    def _start(self):
        if not self._force_async:
            try:
                self._imp = lockfile(
                    self.path, wantnewlockfile=True, flags=os.O_NONBLOCK
                )
            except TryAgain:
                pass
            else:
                self.returncode = os.EX_OK
                self._async_wait()
                return

        if self._force_process or (
            not self._force_thread and self._use_process_by_default
        ):
            self._imp = _LockProcess(path=self.path, scheduler=self.scheduler)
        else:
            self._imp = _LockThread(path=self.path, scheduler=self.scheduler)

        self._imp.addExitListener(self._imp_exit)
        self._imp.start()

    def _imp_exit(self, imp):
        # call exit listeners
        self.returncode = imp.returncode
        self._async_wait()

    def _cancel(self):
        if isinstance(self._imp, AsynchronousTask):
            self._imp.cancel()

    def _poll(self):
        if isinstance(self._imp, AsynchronousTask):
            self._imp.poll()
        return self.returncode

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        if self._imp is None:
            raise AssertionError("not locked")
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        if isinstance(self._imp, (_LockProcess, _LockThread)):
            unlock_future = self._imp.async_unlock()
        else:
            unlockfile(self._imp)
            unlock_future = self.scheduler.create_future()
            self.scheduler.call_soon(unlock_future.set_result, None)
        self._imp = None
        self._unlock_future = unlock_future
        return unlock_future


class _LockThread(AbstractPollTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using a background thread. After the lock is acquired, the thread
    writes to a pipe in order to notify a poll loop running in the main
    thread.
    """

    __slots__ = ("path",) + ("_lock_obj", "_thread", "_unlock_future")

    def _start(self):
        self._registered = True
        threading_mod = threading
        self._thread = threading_mod.Thread(target=self._run_lock)
        self._thread.daemon = True
        self._thread.start()

    def _run_lock(self):
        self._lock_obj = lockfile(self.path, wantnewlockfile=True)
        # Thread-safe callback to EventLoop
        self.scheduler.call_soon_threadsafe(self._run_lock_cb)

    def _run_lock_cb(self):
        self._unregister()
        self.returncode = os.EX_OK
        self._async_wait()

    def _cancel(self):
        # There's currently no way to force thread termination.
        pass

    def _unlock(self):
        if self._lock_obj is None:
            raise AssertionError("not locked")
        if self.returncode is None:
            raise AssertionError("lock not acquired yet")
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        self._unlock_future = self.scheduler.create_future()
        unlockfile(self._lock_obj)
        self._lock_obj = None

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        self._unlock()
        self.scheduler.call_soon(self._unlock_future.set_result, None)
        return self._unlock_future

    def _unregister(self):
        self._registered = False

        if self._thread is not None:
            self._thread.join()
            self._thread = None


class _LockProcess(AbstractPollTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using a subprocess. After the lock is acquired, the process
    writes to a pipe in order to notify a poll loop running in the main
    process. The unlock() method notifies the subprocess to release the
    lock and exit.
    """

    __slots__ = ("path",) + (
        "_acquired",
        "_kill_test",
        "_proc",
        "_files",
        "_unlock_future",
    )

    def _start(self):
        in_pr, in_pw = os.pipe()
        out_pr, out_pw = os.pipe()
        self._files = {}
        self._files["pipe_in"] = in_pr
        self._files["pipe_out"] = out_pw

        fcntl.fcntl(
            in_pr, fcntl.F_SETFL, fcntl.fcntl(in_pr, fcntl.F_GETFL) | os.O_NONBLOCK
        )

        self.scheduler.add_reader(in_pr, self._output_handler)
        self._registered = True
        self._proc = SpawnProcess(
            args=[
                portage._python_interpreter,
                os.path.join(portage._bin_path, "lock-helper.py"),
                self.path,
            ],
            env=dict(os.environ, PORTAGE_PYM_PATH=portage._pym_path),
            fd_pipes={0: out_pr, 1: in_pw, 2: sys.__stderr__.fileno()},
            scheduler=self.scheduler,
        )
        self._proc.addExitListener(self._proc_exit)
        self._proc.start()
        os.close(out_pr)
        os.close(in_pw)

    def _proc_exit(self, proc):
        if self._files is not None:
            # Close pipe_out if it's still open, since it's useless
            # after the process has exited. This helps to avoid
            # "ResourceWarning: unclosed file" since Python 3.2.
            try:
                pipe_out = self._files.pop("pipe_out")
            except KeyError:
                pass
            else:
                os.close(pipe_out)

        if proc.returncode != os.EX_OK:
            # Typically, this will happen due to the
            # process being killed by a signal.

            if not self._acquired:
                # If the lock hasn't been aquired yet, the
                # caller can check the returncode and handle
                # this failure appropriately.
                if not (self.cancelled or self._kill_test):
                    writemsg_level(
                        "_LockProcess: failed to acquire lock on '{self.path}'\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                self._unregister()
                self.returncode = proc.returncode
                self._async_wait()
                return

            if not self.cancelled and self._unlock_future is None:
                # We don't want lost locks going unnoticed, so it's
                # only safe to ignore if either the cancel() or
                # unlock() methods have been previously called.
                raise AssertionError(
                    f"lock process failed with returncode {proc.returncode}"
                )

        if self._unlock_future is not None:
            self._unlock_future.set_result(None)

    def _cancel(self):
        if self._proc is not None:
            self._proc.cancel()

    def _poll(self):
        if self._proc is not None:
            self._proc.poll()
        return self.returncode

    def _output_handler(self):
        buf = self._read_buf(self._files["pipe_in"])
        if buf:
            self._acquired = True
            self._unregister()
            self.returncode = os.EX_OK
            self._async_wait()

        return True

    def _unregister(self):
        self._registered = False

        if self._files is not None:
            try:
                pipe_in = self._files.pop("pipe_in")
            except KeyError:
                pass
            else:
                self.scheduler.remove_reader(pipe_in)
                os.close(pipe_in)

    def _unlock(self):
        if self._proc is None:
            raise AssertionError("not locked")
        if not self._acquired:
            raise AssertionError("lock not acquired yet")
        if self.returncode != os.EX_OK:
            raise AssertionError(
                f"lock process failed with returncode {self.returncode}"
            )
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        self._unlock_future = self.scheduler.create_future()
        os.write(self._files["pipe_out"], b"\0")
        os.close(self._files["pipe_out"])
        self._files = None

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        self._unlock()
        return self._unlock_future



#!/usr/bin/env python
# SOCKSv5 proxy server for network-sandbox
# Copyright 2015-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import asyncio
import errno
import os
import socket
import struct
import sys

if hasattr(asyncio, "ensure_future"):
    # Python >=3.4.4.
    asyncio_ensure_future = asyncio.ensure_future
else:
    # getattr() necessary because async is a keyword in Python >=3.7.
    asyncio_ensure_future = getattr(asyncio, "async")

try:
    current_task = asyncio.current_task
except AttributeError:
    # Deprecated since Python 3.7
    current_task = asyncio.Task.current_task


class Socks5Server:
    """
    An asynchronous SOCKSv5 server.
    """

    async def handle_proxy_conn(self, reader, writer):
        """
        Handle incoming client connection. Perform SOCKSv5 request
        exchange, open a proxied connection and start relaying.

        @param reader: Read side of the socket
        @type reader: asyncio.StreamReader
        @param writer: Write side of the socket
        @type writer: asyncio.StreamWriter
        """

        try:
            # SOCKS hello
            data = await reader.readexactly(2)
            vers, method_no = struct.unpack("!BB", data)

            if vers != 0x05:
                # disconnect on invalid packet -- we have no clue how
                # to reply in alien :)
                writer.close()
                return

            # ...and auth method list
            data = await reader.readexactly(method_no)
            for method in data:
                if method == 0x00:
                    break
            else:
                # no supported method
                method = 0xFF

            # auth reply
            repl = struct.pack("!BB", 0x05, method)
            writer.write(repl)
            await writer.drain()
            if method == 0xFF:
                writer.close()
                return

            # request
            data = await reader.readexactly(4)
            vers, cmd, rsv, atyp = struct.unpack("!BBBB", data)

            if vers != 0x05 or rsv != 0x00:
                # disconnect on malformed packet
                self.close()
                return

            # figure out if we can handle it
            rpl = 0x00
            if cmd != 0x01:  # CONNECT
                rpl = 0x07  # command not supported
            elif atyp == 0x01:  # IPv4
                data = await reader.readexactly(4)
                addr = socket.inet_ntoa(data)
            elif atyp == 0x03:  # domain name
                data = await reader.readexactly(1)
                (addr_len,) = struct.unpack("!B", data)
                addr = await reader.readexactly(addr_len)
                try:
                    addr = addr.decode("idna")
                except UnicodeDecodeError:
                    rpl = 0x04  # host unreachable

            elif atyp == 0x04:  # IPv6
                data = await reader.readexactly(16)
                addr = socket.inet_ntop(socket.AF_INET6, data)
            else:
                rpl = 0x08  # address type not supported

            # try to connect if we can handle it
            if rpl == 0x00:
                data = await reader.readexactly(2)
                (port,) = struct.unpack("!H", data)

                try:
                    # open a proxied connection
                    proxied_reader, proxied_writer = await asyncio.open_connection(
                        addr, port
                    )
                except (socket.gaierror, socket.herror):
                    # DNS failure
                    rpl = 0x04  # host unreachable
                except OSError as e:
                    # connection failure
                    if e.errno in (errno.ENETUNREACH, errno.ENETDOWN):
                        rpl = 0x03  # network unreachable
                    elif e.errno in (errno.EHOSTUNREACH, errno.EHOSTDOWN):
                        rpl = 0x04  # host unreachable
                    elif e.errno in (errno.ECONNREFUSED, errno.ETIMEDOUT):
                        rpl = 0x05  # connection refused
                    else:
                        raise
                else:
                    # get socket details that we can send back to the client
                    # local address (sockname) in particular -- but we need
                    # to ask for the whole socket since Python's sockaddr
                    # does not list the family...
                    sock = proxied_writer.get_extra_info("socket")
                    addr = sock.getsockname()
                    if sock.family == socket.AF_INET:
                        host, port = addr
                        bin_host = socket.inet_aton(host)

                        repl_addr = struct.pack("!B4sH", 0x01, bin_host, port)
                    elif sock.family == socket.AF_INET6:
                        # discard flowinfo, scope_id
                        host, port = addr[:2]
                        bin_host = socket.inet_pton(sock.family, host)

                        repl_addr = struct.pack("!B16sH", 0x04, bin_host, port)

            if rpl != 0x00:
                # fallback to 0.0.0.0:0
                repl_addr = struct.pack("!BLH", 0x01, 0x00000000, 0x0000)

            # reply to the request
            repl = struct.pack("!BBB", 0x05, rpl, 0x00)
            writer.write(repl + repl_addr)
            await writer.drain()

            # close if an error occured
            if rpl != 0x00:
                writer.close()
                return

            # otherwise, start two loops:
            # remote -> local...
            t = asyncio_ensure_future(
                self.handle_proxied_conn(proxied_reader, writer, current_task())
            )

            # and local -> remote...
            try:
                try:
                    while True:
                        data = await reader.read(4096)
                        if data == b"":
                            # client disconnected, stop relaying from
                            # remote host
                            t.cancel()
                            break

                        proxied_writer.write(data)
                        await proxied_writer.drain()
                except OSError:
                    # read or write failure
                    t.cancel()
                except:
                    t.cancel()
                    raise
            finally:
                # always disconnect in the end :)
                proxied_writer.close()
                writer.close()

        except (OSError, asyncio.IncompleteReadError, asyncio.CancelledError):
            writer.close()
            return
        except:
            writer.close()
            raise

    async def handle_proxied_conn(self, proxied_reader, writer, parent_task):
        """
        Handle the proxied connection. Relay incoming data
        to the client.

        @param reader: Read side of the socket
        @type reader: asyncio.StreamReader
        @param writer: Write side of the socket
        @type writer: asyncio.StreamWriter
        """

        try:
            try:
                while True:
                    data = await proxied_reader.read(4096)
                    if data == b"":
                        break

                    writer.write(data)
                    await writer.drain()
            finally:
                parent_task.cancel()
        except (OSError, asyncio.CancelledError):
            return


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <socket-path>")
        sys.exit(1)

    loop = asyncio.new_event_loop()
    s = Socks5Server()
    server = loop.run_until_complete(
        asyncio.start_unix_server(s.handle_proxy_conn, sys.argv[1])
    )

    ret = 0
    try:
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        except:
            ret = 1
    finally:
        server.close()
        loop.run_until_complete(server.wait_closed())
        loop.close()
        os.unlink(sys.argv[1])



# testCommandOututSet.py -- Portage Unit Testing Functionality
# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.process import find_binary
from portage.tests import TestCase, test_cps
from portage._sets.shell import CommandOutputSet


class CommandOutputSetTestCase(TestCase):
    """Simple Test Case for CommandOutputSet"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testCommand(self):
        params = set(test_cps)
        command = find_binary("bash")
        command += " -c '"
        for a in params:
            command += f' echo -e "{a}" ; '
        command += "'"
        s = CommandOutputSet(command)
        atoms = s.getAtoms()
        self.assertEqual(atoms, params)



#!/usr/bin/env python
# Copyright 2010-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
#
# This is a helper which ebuild processes can use
# to communicate with portage's main python process.

import os
import signal

# For compatibility with Python < 3.8
raise_signal = getattr(
    signal, "raise_signal", lambda signum: os.kill(os.getpid(), signum)
)


# Inherit from KeyboardInterrupt to avoid a traceback from asyncio.
class SignalInterrupt(KeyboardInterrupt):
    def __init__(self, signum):
        self.signum = signum


try:

    def signal_interrupt(signum, _frame):
        raise SignalInterrupt(signum)

    def debug_signal(_signum, _frame):
        import pdb

        pdb.set_trace()

    # Prevent "[Errno 32] Broken pipe" exceptions when writing to a pipe.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal_interrupt)
    signal.signal(signal.SIGUSR1, debug_signal)

    import errno
    import logging
    import pickle
    import sys
    import time

    if os.path.isfile(
        os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
            ".portage_not_installed",
        )
    ):
        pym_paths = [
            os.path.join(
                os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "lib"
            )
        ]
        sys.path.insert(0, pym_paths[0])
    else:
        import sysconfig

        pym_paths = [
            os.path.join(sysconfig.get_path("purelib"), x)
            for x in ("_emerge", "portage")
        ]
    # Avoid sandbox violations after Python upgrade.
    if os.environ.get("SANDBOX_ON") == "1":
        sandbox_write = os.environ.get("SANDBOX_WRITE", "").split(":")
        for pym_path in pym_paths:
            if pym_path not in sandbox_write:
                sandbox_write.append(pym_path)
                os.environ["SANDBOX_WRITE"] = ":".join(filter(None, sandbox_write))
        del pym_path, sandbox_write
    del pym_paths

    import portage

    portage._internal_caller = True
    portage._disable_legacy_globals()

    from portage.util._eventloop.global_event_loop import global_event_loop
    from _emerge.AbstractPollTask import AbstractPollTask
    from _emerge.PipeReader import PipeReader

    RETURNCODE_WRITE_FAILED = 2

    class FifoWriter(AbstractPollTask):
        __slots__ = ("buf", "fifo", "_fd")

        def _start(self):
            try:
                self._fd = os.open(self.fifo, os.O_WRONLY | os.O_NONBLOCK)
            except OSError as e:
                if e.errno == errno.ENXIO:
                    # This happens if the daemon has been killed.
                    self.returncode = RETURNCODE_WRITE_FAILED
                    self._unregister()
                    self._async_wait()
                    return
                else:
                    raise
            self.scheduler.add_writer(self._fd, self._output_handler)
            self._registered = True

        def _output_handler(self):
            # The whole buf should be able to fit in the fifo with
            # a single write call, so there's no valid reason for
            # os.write to raise EAGAIN here.
            fd = self._fd
            buf = self.buf
            while buf:
                try:
                    buf = buf[os.write(fd, buf) :]
                except OSError:
                    self.returncode = RETURNCODE_WRITE_FAILED
                    self._async_wait()
                    return

            self.returncode = os.EX_OK
            self._async_wait()

        def _cancel(self):
            self.returncode = self._cancelled_returncode
            self._unregister()

        def _unregister(self):
            self._registered = False
            if self._fd is not None:
                self.scheduler.remove_writer(self._fd)
                os.close(self._fd)
                self._fd = None

    class EbuildIpc:
        # Timeout for each individual communication attempt (we retry
        # as long as the daemon process appears to be alive).
        _COMMUNICATE_RETRY_TIMEOUT = 15  # seconds

        def __init__(self):
            self.fifo_dir = os.environ["PORTAGE_BUILDDIR"]
            self.ipc_in_fifo = os.path.join(self.fifo_dir, ".ipc", "in")
            self.ipc_out_fifo = os.path.join(self.fifo_dir, ".ipc", "out")
            self.ipc_lock_file = os.path.join(self.fifo_dir, ".ipc", "lock")

        def _daemon_is_alive(self):
            try:
                builddir_lock = portage.locks.lockfile(
                    self.fifo_dir, wantnewlockfile=True, flags=os.O_NONBLOCK
                )
            except portage.exception.TryAgain:
                return True
            else:
                portage.locks.unlockfile(builddir_lock)
                return False

        def communicate(self, args):
            # Make locks quiet since unintended locking messages displayed on
            # stdout could corrupt the intended output of this program.
            portage.locks._quiet = True
            lock_obj = portage.locks.lockfile(self.ipc_lock_file, unlinkfile=True)

            try:
                return self._communicate(args)
            finally:
                portage.locks.unlockfile(lock_obj)

        def _timeout_retry_msg(self, start_time, when):
            time_elapsed = time.time() - start_time
            portage.util.writemsg_level(
                f"ebuild-ipc timed out {when} after {time_elapsed} seconds, retrying...\n",
                level=logging.ERROR,
                noiselevel=-1,
            )

        def _no_daemon_msg(self):
            portage.util.writemsg_level(
                portage.localization._("ebuild-ipc: daemon process not detected\n"),
                level=logging.ERROR,
                noiselevel=-1,
            )

        def _run_writer(self, fifo_writer, msg):
            """
            Wait on pid and return an appropriate exit code. This
            may return unsuccessfully due to timeout if the daemon
            process does not appear to be alive.
            """

            start_time = time.time()

            fifo_writer.start()
            eof = fifo_writer.poll() is not None

            while not eof:
                fifo_writer._wait_loop(timeout=self._COMMUNICATE_RETRY_TIMEOUT)

                eof = fifo_writer.poll() is not None
                if eof:
                    break
                elif self._daemon_is_alive():
                    self._timeout_retry_msg(start_time, msg)
                else:
                    fifo_writer.cancel()
                    self._no_daemon_msg()
                    fifo_writer.wait()
                    return 2

            return fifo_writer.wait()

        def _receive_reply(self, input_fd):
            start_time = time.time()

            pipe_reader = PipeReader(
                input_files={"input_fd": input_fd}, scheduler=global_event_loop()
            )
            pipe_reader.start()

            eof = pipe_reader.poll() is not None

            while not eof:
                pipe_reader._wait_loop(timeout=self._COMMUNICATE_RETRY_TIMEOUT)
                eof = pipe_reader.poll() is not None
                if not eof:
                    if self._daemon_is_alive():
                        self._timeout_retry_msg(
                            start_time, portage.localization._("during read")
                        )
                    else:
                        pipe_reader.cancel()
                        self._no_daemon_msg()
                        return 2

            buf = pipe_reader.getvalue()

            retval = 2

            if not buf:
                portage.util.writemsg_level(
                    f"ebuild-ipc: {portage.localization._('read failed')}\n",
                    level=logging.ERROR,
                    noiselevel=-1,
                )

            else:
                try:
                    reply = pickle.loads(buf)
                except SystemExit:
                    raise
                except Exception as e:
                    # The pickle module can raise practically
                    # any exception when given corrupt data.
                    portage.util.writemsg_level(
                        f"ebuild-ipc: {e}\n", level=logging.ERROR, noiselevel=-1
                    )

                else:
                    (out, err, retval) = reply

                    if out:
                        portage.util.writemsg_stdout(out, noiselevel=-1)

                    if err:
                        portage.util.writemsg(err, noiselevel=-1)

            return retval

        def _communicate(self, args):
            if not self._daemon_is_alive():
                self._no_daemon_msg()
                return 2

            # Open the input fifo before the output fifo, in order to make it
            # possible for the daemon to send a reply without blocking. This
            # improves performance, and also makes it possible for the daemon
            # to do a non-blocking write without a race condition.
            input_fd = os.open(self.ipc_out_fifo, os.O_RDONLY | os.O_NONBLOCK)

            # Use forks so that the child process can handle blocking IO
            # un-interrupted, while the parent handles all timeout
            # considerations. This helps to avoid possible race conditions
            # from interference between timeouts and blocking IO operations.
            msg = portage.localization._("during write")
            retval = self._run_writer(
                FifoWriter(
                    buf=pickle.dumps(args),
                    fifo=self.ipc_in_fifo,
                    scheduler=global_event_loop(),
                ),
                msg,
            )

            if retval != os.EX_OK:
                portage.util.writemsg_level(
                    f"ebuild-ipc: {msg}: subprocess failure: {retval}\n",
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                return retval

            if not self._daemon_is_alive():
                self._no_daemon_msg()
                return 2

            return self._receive_reply(input_fd)

    def ebuild_ipc_main(args):
        ebuild_ipc = EbuildIpc()
        return ebuild_ipc.communicate(args)

    if __name__ == "__main__":
        try:
            sys.exit(ebuild_ipc_main(sys.argv[1:]))
        finally:
            global_event_loop().close()

except KeyboardInterrupt as e:
    # Prevent traceback on ^C
    signum = getattr(e, "signum", signal.SIGINT)
    signal.signal(signum, signal.SIG_DFL)
    raise_signal(signum)



# Copyright 2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from _emerge.unmerge import _unmerge_display

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground


class _TestData:
    def __init__(self, unmerge_files, expected_pkgmap):
        self.unmerge_files = unmerge_files

        # The pkgmap created by unmerge_display is a list where each entry is of the form
        # {"selected": list(...), "omitted": set(...), "protected": set(...) }.
        # To simplify the notation of the test data, we receive a list with entries of the form
        # (s1,o1)
        # The entries are then translated to the expected form:
        # {"selected": s1, "omitted": o1, "protected": set()}
        # The "protected" field is not relevant for testing ordering.
        # The ordering of the "omitted" field is not relevant.
        expand = lambda x: {"selected": x[0], "omitted": set(x[1]), "protected": set()}
        self.expected_pkgmap = list(map(expand, expected_pkgmap))


class UnmergeOrderTestCase(TestCase):
    def testUnmergeOrder(self):
        ebuilds = {
            "c/x-1": {},
            "c/y-2": {},
            "c/y-3": {},
            "c/z-4": {},
            "c/z-5": {},
            "c/z-6": {},
            "c/zz-4": {},
            "c/zz-5": {},
            "c/zz-6": {},
        }
        installed = {
            "c/x-1": {},
            "c/y-2": {},
            "c/z-4": {},
            "c/z-5": {},
            "c/z-6": {},
            "c/zz-4": {},
            "c/zz-5": {},
            "c/zz-6": {},
        }
        test_cases = (
            # cp = category/package
            # cpv = category/package-version
            # Single cpv atom, representing the only available instance of the cp.
            # The pkgmap should contain exactly that cpv and no omitted packages.
            _TestData(["c/x-1"], [(["c/x-1"], [])]),
            # Single cp atom. The pkgmap should contain the only available cpv to
            # which the cp expands, no omitted packages.
            _TestData(["c/x"], [(["c/x-1"], [])]),
            # Duplicate cpv atom, representing the only available instance of the cp.
            # The pkgmap should contain the cpv with no omitted packages, and an empty
            # entry representing the duplicate.
            _TestData(["c/x-1", "c/x-1"], [(["c/x-1"], []), ([], [])]),
            # Duplicate cp atom, representing the only available instance. The pkgmap
            # should contain the only available cpv to which the cp expands, with no
            # omitted packages, and a second empty entry representing the duplicate.
            _TestData(["c/x", "c/x"], [(["c/x-1"], []), ([], [])]),
            # Single cpv atom, representing one of the two available instances. The
            # pkgmap should contain exactly that cpv. Since the other instance is not
            # installed, there should be no omitted packages.
            _TestData(["c/y-2"], [(["c/y-2"], [])]),
            # Single cp atom. The pkgmap should contain exactly the only installed
            # instance and no omitted packages.
            _TestData(["c/y"], [(["c/y-2"], [])]),
            # Single cpv atom, representing one of the three available instances.
            # The pkgmap should contain exactly the cpv. Since all three instances
            # are installed, the other two instances should be in the omitted packages.
            _TestData(["c/z-4"], [(["c/z-4"], ["c/z-5", "c/z-6"])]),
            # Single cp atom. The pkgmap should contain all three installed instances.
            # Since there are no other installed instances, there should be no omitted
            # packages.
            _TestData(["c/z"], [(["c/z-4", "c/z-5", "c/z-6"], [])]),
            # Two cpv atoms belonging to the same cp. The pkgmap should contain an
            # entry for each cpv, in the same order. The third installed cpv belonging
            # to the cp should be listed in the omitted section of each entry.
            _TestData(
                ["c/z-4", "c/z-5"], [(["c/z-4"], ["c/z-6"]), (["c/z-5"], ["c/z-6"])]
            ),
            _TestData(
                ["c/z-5", "c/z-4"], [(["c/z-5"], ["c/z-6"]), (["c/z-4"], ["c/z-6"])]
            ),
            # Three cpv atoms belonging to the same cp. The pkgmap should contain an
            # entry for each cpv, in the same order. Since there are no other instances
            # of the cp, the omitted section of each entry should be empty.
            _TestData(
                ["c/z-4", "c/z-5", "c/z-6"],
                [(["c/z-4"], []), (["c/z-5"], []), (["c/z-6"], [])],
            ),
            _TestData(
                ["c/z-6", "c/z-5", "c/z-4"],
                [(["c/z-6"], []), (["c/z-5"], []), (["c/z-4"], [])],
            ),
            # First a cp atom, then a cpv atom that is an instance of the cp. The
            # pkgmap should contain an entry containing all installed cpv's that the cp
            # expands to, in sorted order. It should then contain an empty entry
            # representing the input cpv that is already covered by the expansion of
            # the cp.
            _TestData(["c/z", "c/z-4"], [(["c/z-4", "c/z-5", "c/z-6"], []), ([], [])]),
            _TestData(["c/z", "c/z-6"], [(["c/z-4", "c/z-5", "c/z-6"], []), ([], [])]),
            # First a cpv atom, then the cp to which the cpv belongs. The pkgmap
            # should contain an entry for the first cpv, then an entry containing
            # the remaining cpv's to which the cp expands.
            _TestData(["c/z-4", "c/z"], [(["c/z-4"], []), (["c/z-5", "c/z-6"], [])]),
            _TestData(["c/z-6", "c/z"], [(["c/z-6"], []), (["c/z-4", "c/z-5"], [])]),
            # More mixed cp/cpv's. The cp should expand to all cpv's except those
            # covered by a preceding cpv. The cpv's after the cp should result in empty
            # entries, since they are already covered by the expansion of the cp.
            _TestData(
                ["c/z", "c/z-4", "c/z-5"],
                [(["c/z-4", "c/z-5", "c/z-6"], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z", "c/z-5", "c/z-4"],
                [(["c/z-4", "c/z-5", "c/z-6"], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z-4", "c/z", "c/z-5"],
                [(["c/z-4"], []), (["c/z-5", "c/z-6"], []), ([], [])],
            ),
            _TestData(
                ["c/z-5", "c/z", "c/z-4"],
                [(["c/z-5"], []), (["c/z-4", "c/z-6"], []), ([], [])],
            ),
            _TestData(
                ["c/z-4", "c/z-5", "c/z"],
                [(["c/z-4"], []), (["c/z-5"], []), (["c/z-6"], [])],
            ),
            _TestData(
                ["c/z-5", "c/z-4", "c/z"],
                [(["c/z-5"], []), (["c/z-4"], []), (["c/z-6"], [])],
            ),
            _TestData(
                ["c/z", "c/z-4", "c/z-5", "c/z-6"],
                [(["c/z-4", "c/z-5", "c/z-6"], []), ([], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z", "c/z-6", "c/z-5", "c/z-4"],
                [(["c/z-4", "c/z-5", "c/z-6"], []), ([], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z-4", "c/z", "c/z-5", "c/z-6"],
                [(["c/z-4"], []), (["c/z-5", "c/z-6"], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z-6", "c/z", "c/z-5", "c/z-4"],
                [(["c/z-6"], []), (["c/z-4", "c/z-5"], []), ([], []), ([], [])],
            ),
            _TestData(
                ["c/z-4", "c/z-5", "c/z", "c/z-6"],
                [(["c/z-4"], []), (["c/z-5"], []), (["c/z-6"], []), ([], [])],
            ),
            _TestData(
                ["c/z-6", "c/z-5", "c/z", "c/z-4"],
                [(["c/z-6"], []), (["c/z-5"], []), (["c/z-4"], []), ([], [])],
            ),
            _TestData(
                ["c/z-4", "c/z-5", "c/z-6", "c/z"],
                [(["c/z-4"], []), (["c/z-5"], []), (["c/z-6"], []), ([], [])],
            ),
            _TestData(
                ["c/z-6", "c/z-5", "c/z-4", "c/z"],
                [(["c/z-6"], []), (["c/z-5"], []), (["c/z-4"], []), ([], [])],
            ),
            # Two cpv that do not belong to the same cp. The pkgmap should contain an
            # entry for each cpv, in the same order. If there are other installed
            # instances of the cp to which the cpv belongs, they should be listed
            # in the omitted section.
            _TestData(["c/x-1", "c/y-2"], [(["c/x-1"], []), (["c/y-2"], [])]),
            _TestData(["c/y-2", "c/x-1"], [(["c/y-2"], []), (["c/x-1"], [])]),
            _TestData(
                ["c/x-1", "c/z-4"], [(["c/x-1"], []), (["c/z-4"], ["c/z-5", "c/z-6"])]
            ),
            _TestData(
                ["c/z-4", "c/x-1"], [(["c/z-4"], ["c/z-5", "c/z-6"]), (["c/x-1"], [])]
            ),
            # cpv's/cp where some cpv's are not instances of the cp. The pkgmap should
            # contain an entry for each in the same order, with the cp expanded
            # to all installed instances.
            _TestData(
                ["c/x-1", "c/z"], [(["c/x-1"], []), (["c/z-4", "c/z-5", "c/z-6"], [])]
            ),
            _TestData(
                ["c/z", "c/x-1"], [(["c/z-4", "c/z-5", "c/z-6"], []), (["c/x-1"], [])]
            ),
            _TestData(
                ["c/x-1", "c/z-4", "c/z"],
                [(["c/x-1"], []), (["c/z-4"], []), (["c/z-5", "c/z-6"], [])],
            ),
            _TestData(
                ["c/z-4", "c/z", "c/x-1"],
                [(["c/z-4"], []), (["c/z-5", "c/z-6"], []), (["c/x-1"], [])],
            ),
            _TestData(
                ["c/x-1", "c/z", "c/z-4"],
                [(["c/x-1"], []), (["c/z-4", "c/z-5", "c/z-6"], []), ([], [])],
            ),
            _TestData(
                ["c/z", "c/z-4", "c/x-1"],
                [(["c/z-4", "c/z-5", "c/z-6"], []), ([], []), (["c/x-1"], [])],
            ),
            # Two different cp's. The pkglist should contain an entry for each cp,
            # in the same order, containing all cpv's that the cp's expands to.
            _TestData(
                ["c/z", "c/zz"],
                [
                    (["c/z-4", "c/z-5", "c/z-6"], []),
                    (["c/zz-4", "c/zz-5", "c/zz-6"], []),
                ],
            ),
            _TestData(
                ["c/zz", "c/z"],
                [
                    (["c/zz-4", "c/zz-5", "c/zz-6"], []),
                    (["c/z-4", "c/z-5", "c/z-6"], []),
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)

        try:
            for test_case in test_cases:
                eroot = playground.settings["EROOT"]
                root_config = playground.trees[eroot]["root_config"]

                res, pkgmap = _unmerge_display(
                    root_config, [], "unmerge", test_case.unmerge_files, ordered=True
                )

                self.assertEqual(res, os.EX_OK)
                self.assertEqual(pkgmap, test_case.expected_pkgmap)
        finally:
            playground.cleanup()



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

"""
Base class for performing repos.conf sync variables checks.
This class contains common checks code and functions.

For additional checks or other customizations,
subclass it adding and/or overriding classes as needed.
"""

import logging

from portage.localization import _
from portage.util import writemsg_level


def check_type(repo, logger, module_names):
    if repo.sync_uri is not None and repo.sync_type is None:
        writemsg_level(
            "!!! %s\n"
            % _(
                "Repository '%s' has sync-uri attribute, but is missing sync-type attribute"
            )
            % repo.name,
            level=logger.ERROR,
            noiselevel=-1,
        )
        return False
    if repo.sync_type not in module_names + [None]:
        writemsg_level(
            "!!! %s\n"
            % _(
                "Repository '%s' has sync-type attribute set to unsupported value: '%s'"
            )
            % (repo.name, repo.sync_type),
            level=logger.ERROR,
            noiselevel=-1,
        )
        writemsg_level(
            "!!! %s\n" % _("Installed sync-types are: '%s'") % (str(module_names)),
            level=logger.ERROR,
            noiselevel=-1,
        )
        return False
    return True


class CheckSyncConfig:
    """Base repos.conf settings checks class"""

    def __init__(self, repo=None, logger=None):
        """Class init function

        @param logger: optional logging instance,
                defaults to logging module
        """
        self.logger = logger or logging
        self.repo = repo
        self.checks = ["check_uri", "check_auto_sync"]

    def repo_checks(self):
        """Perform all checks available"""
        for check in self.checks:
            getattr(self, check)()

    def check_uri(self):
        """Check the sync_uri setting"""
        if self.repo.sync_uri is None:
            writemsg_level(
                "!!! %s\n"
                % _(
                    "Repository '%s' has sync-type attribute, but is missing sync-uri attribute"
                )
                % self.repo.name,
                level=self.logger.ERROR,
                noiselevel=-1,
            )

    def check_auto_sync(self):
        """Check the auto_sync setting"""
        if self.repo.auto_sync is None:
            writemsg_level(
                "!!! %s\n"
                % _("Repository '%s' is missing auto_sync attribute")
                % self.repo.name,
                level=self.logger.ERROR,
                noiselevel=-1,
            )
        elif self.repo.auto_sync.lower() not in ["yes", "true", "no", "false"]:
            writemsg_level(
                "!!! %s\n"
                % _("Repository '%s' auto_sync attribute must be one of: %s")
                % (self.repo.name, "{yes, true, no, false}"),
                level=self.logger.ERROR,
                noiselevel=-1,
            )



# Copyright 2018-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from portage.tests import TestCase
from portage.util._async.AsyncFunction import AsyncFunction
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures.unix_events import DefaultEventLoopPolicy


def fork_main():
    loop = asyncio._wrap_loop()
    # This fails with python's default event loop policy,
    # see https://bugs.python.org/issue22087.
    loop.run_until_complete(asyncio.sleep(0.1, loop=loop))
    loop.close()


def async_main(fork_exitcode, loop=None):
    loop = asyncio._wrap_loop(loop)
    proc = AsyncFunction(scheduler=loop, target=fork_main)
    proc.start()
    proc.async_wait().add_done_callback(
        lambda future: fork_exitcode.set_result(future.result())
    )


class EventLoopInForkTestCase(TestCase):
    """
    The default asyncio event loop policy does not support loops
    running in forks, see https://bugs.python.org/issue22087.
    Portage's DefaultEventLoopPolicy supports forks.
    """

    def testEventLoopInForkTestCase(self):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())
        loop = None
        try:
            loop = asyncio._wrap_loop()
            fork_exitcode = loop.create_future()
            # Make async_main fork while the loop is running, which would
            # trigger https://bugs.python.org/issue22087 with asyncio's
            # default event loop policy.
            loop.call_soon(async_main, fork_exitcode)
            assert loop.run_until_complete(fork_exitcode) == os.EX_OK
        finally:
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures.unix_events import DefaultEventLoopPolicy


class RunUntilCompleteTestCase(TestCase):
    def test_add_done_callback(self):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        loop = None
        try:
            loop = asyncio._wrap_loop()
            f1 = loop.create_future()
            f2 = loop.create_future()
            f1.add_done_callback(f2.set_result)
            loop.call_soon(lambda: f1.set_result(None))
            loop.run_until_complete(f1)
            self.assertEqual(f1.done(), True)

            # This proves that done callbacks of f1 are executed before
            # loop.run_until_complete(f1) returns, which is how asyncio's
            # default event loop behaves.
            self.assertEqual(f2.done(), True)
        finally:
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())



# Copyright 2018-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import pty
import shutil
import socket
import tempfile

from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures.unix_events import (
    DefaultEventLoopPolicy,
    _set_nonblocking,
)


class _PipeClosedTestCase:
    def test_pipe(self):
        read_end, write_end = os.pipe()
        self._do_test(read_end, write_end)

    def test_pty_device(self):
        try:
            read_end, write_end = pty.openpty()
        except OSError:
            self.skipTest("pty not available")
        self._do_test(read_end, write_end)

    def test_domain_socket(self):
        read_end, write_end = socket.socketpair()
        self._do_test(read_end.detach(), write_end.detach())

    def test_named_pipe(self):
        tempdir = tempfile.mkdtemp()
        try:
            fifo_path = os.path.join(tempdir, "fifo")
            os.mkfifo(fifo_path)
            self._do_test(
                os.open(fifo_path, os.O_NONBLOCK | os.O_RDONLY),
                os.open(fifo_path, os.O_NONBLOCK | os.O_WRONLY),
            )
        finally:
            shutil.rmtree(tempdir)


class ReaderPipeClosedTestCase(_PipeClosedTestCase, TestCase):
    """
    Test that a reader callback is called after the other end of
    the pipe has been closed.
    """

    def _do_test(self, read_end, write_end):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        loop = asyncio._wrap_loop()
        read_end = os.fdopen(read_end, "rb", 0)
        write_end = os.fdopen(write_end, "wb", 0)
        try:

            def reader_callback():
                if not reader_callback.called.done():
                    reader_callback.called.set_result(None)

            reader_callback.called = loop.create_future()
            loop.add_reader(read_end.fileno(), reader_callback)

            # Allow the loop to check for IO events, and assert
            # that our future is still not done.
            loop.run_until_complete(asyncio.sleep(0, loop=loop))
            self.assertFalse(reader_callback.called.done())

            # Demonstrate that the callback is called after the
            # other end of the pipe has been closed.
            write_end.close()
            loop.run_until_complete(reader_callback.called)
        finally:
            loop.remove_reader(read_end.fileno())
            write_end.close()
            read_end.close()
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())


class WriterPipeClosedTestCase(_PipeClosedTestCase, TestCase):
    """
    Test that a writer callback is called after the other end of
    the pipe has been closed.
    """

    def _do_test(self, read_end, write_end):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        loop = asyncio._wrap_loop()
        read_end = os.fdopen(read_end, "rb", 0)
        write_end = os.fdopen(write_end, "wb", 0)
        try:

            def writer_callback():
                if not writer_callback.called.done():
                    writer_callback.called.set_result(None)

            writer_callback.called = loop.create_future()
            _set_nonblocking(write_end.fileno())
            loop.add_writer(write_end.fileno(), writer_callback)

            # With pypy we've seen intermittent spurious writer callbacks
            # here, so retry until the correct state is achieved.
            tries = 10
            while tries:
                tries -= 1

                # Fill up the pipe, so that no writer callbacks should be
                # received until the state has changed.
                while True:
                    try:
                        os.write(write_end.fileno(), 512 * b"0")
                    except OSError as e:
                        if e.errno != errno.EAGAIN:
                            raise
                        break

                # Allow the loop to check for IO events, and assert
                # that our future is still not done.
                loop.run_until_complete(asyncio.sleep(0, loop=loop))
                if writer_callback.called.done():
                    writer_callback.called = loop.create_future()
                else:
                    break

            self.assertFalse(writer_callback.called.done())

            # Demonstrate that the callback is called after the
            # other end of the pipe has been closed.
            read_end.close()
            loop.run_until_complete(writer_callback.called)
        finally:
            loop.remove_writer(write_end.fileno())
            write_end.close()
            read_end.close()
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())



# cvstree.py -- cvs tree utilities
# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import re
import stat
import time

from portage import os
from portage import _encodings
from portage import _unicode_encode


# [D]/Name/Version/Date/Flags/Tags


def pathdata(entries, path):
    """Returns the data(dict) for a specific file/dir at the path specified."""
    mysplit = path.split("/")
    myentries = entries
    mytarget = mysplit[-1]
    mysplit = mysplit[:-1]
    for mys in mysplit:
        if mys in myentries["dirs"]:
            myentries = myentries["dirs"][mys]
        else:
            return None
    if mytarget in myentries["dirs"]:
        return myentries["dirs"][mytarget]
    if mytarget in myentries["files"]:
        return myentries["files"][mytarget]
    return None


def fileat(entries, path):
    return pathdata(entries, path)


def isadded(entries, path):
    """Returns True if the path exists and is added to the cvs tree."""
    mytarget = pathdata(entries, path)
    if mytarget:
        if "cvs" in mytarget["status"]:
            return 1

    basedir = os.path.dirname(path)
    filename = os.path.basename(path)

    try:
        myfile = open(
            _unicode_encode(
                os.path.join(basedir, "CVS", "Entries"),
                encoding=_encodings["fs"],
                errors="strict",
            ),
            encoding=_encodings["content"],
            errors="strict",
        )
    except OSError:
        return 0
    mylines = myfile.readlines()
    myfile.close()

    rep = re.compile(r"^\/%s\/" % re.escape(filename))
    for x in mylines:
        if rep.search(x):
            return 1

    return 0


def findnew(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all elements that have been added but
    have not yet been committed. Returns a list of paths, optionally prepended
    with a basedir.
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"

    mylist = []
    for myfile in entries["files"]:
        if "cvs" in entries["files"][myfile]["status"]:
            if "0" == entries["files"][myfile]["revision"]:
                mylist.append(basedir + myfile)

    if recursive:
        for mydir in entries["dirs"]:
            mylist += findnew(entries["dirs"][mydir], recursive, basedir + mydir)

    return mylist


def findoption(entries, pattern, recursive=0, basedir=""):
    """Iterate over paths of cvs entries for which the pattern.search() method
    finds a match. Returns a list of paths, optionally prepended with a
    basedir.
    """
    if not basedir.endswith("/"):
        basedir += "/"

    for myfile, mydata in entries["files"].items():
        if "cvs" in mydata["status"]:
            if pattern.search(mydata["flags"]):
                yield basedir + myfile

    if recursive:
        for mydir, mydata in entries["dirs"].items():
            yield from findoption(mydata, pattern, recursive, basedir + mydir)


def findchanged(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all elements that exist in the cvs tree
    and differ from the committed version. Returns a list of paths, optionally
    prepended with a basedir.
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"

    mylist = []
    for myfile in entries["files"]:
        if "cvs" in entries["files"][myfile]["status"]:
            if "current" not in entries["files"][myfile]["status"]:
                if "exists" in entries["files"][myfile]["status"]:
                    if entries["files"][myfile]["revision"] != "0":
                        mylist.append(basedir + myfile)

    if recursive:
        for mydir in entries["dirs"]:
            mylist += findchanged(entries["dirs"][mydir], recursive, basedir + mydir)

    return mylist


def findmissing(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all elements that are listed in the cvs
    tree but do not exist on the filesystem. Returns a list of paths,
    optionally prepended with a basedir.
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"

    mylist = []
    for myfile in entries["files"]:
        if "cvs" in entries["files"][myfile]["status"]:
            if "exists" not in entries["files"][myfile]["status"]:
                if "removed" not in entries["files"][myfile]["status"]:
                    mylist.append(basedir + myfile)

    if recursive:
        for mydir in entries["dirs"]:
            mylist += findmissing(entries["dirs"][mydir], recursive, basedir + mydir)

    return mylist


def findunadded(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all elements that are in valid cvs
    directories but are not part of the cvs tree. Returns a list of paths,
    optionally prepended with a basedir.
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"

    # Ignore what cvs ignores.
    mylist = []
    for myfile in entries["files"]:
        if "cvs" not in entries["files"][myfile]["status"]:
            mylist.append(basedir + myfile)

    if recursive:
        for mydir in entries["dirs"]:
            mylist += findunadded(entries["dirs"][mydir], recursive, basedir + mydir)

    return mylist


def findremoved(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all elements that are in flagged for cvs
    deletions. Returns a list of paths,	optionally prepended with a basedir.
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"

    mylist = []
    for myfile in entries["files"]:
        if "removed" in entries["files"][myfile]["status"]:
            mylist.append(basedir + myfile)

    if recursive:
        for mydir in entries["dirs"]:
            mylist += findremoved(entries["dirs"][mydir], recursive, basedir + mydir)

    return mylist


def findall(entries, recursive=0, basedir=""):
    """Recurses the entries tree to find all new, changed, missing, and unadded
    entities. Returns a 4 element list of lists as returned from each find*().
    """
    if basedir and basedir[-1] != "/":
        basedir += "/"
    mynew = findnew(entries, recursive, basedir)
    mychanged = findchanged(entries, recursive, basedir)
    mymissing = findmissing(entries, recursive, basedir)
    myunadded = findunadded(entries, recursive, basedir)
    myremoved = findremoved(entries, recursive, basedir)
    return [mynew, mychanged, mymissing, myunadded, myremoved]


ignore_list = re.compile(
    r"(^|/)(RCS(|LOG)|SCCS|CVS(|\.adm)|cvslog\..*|tags|TAGS|\.(make\.state|nse_depinfo)|.*~|(\.|)#.*|,.*|_$.*|.*\$|\.del-.*|.*\.(old|BAK|bak|orig|rej|a|olb|o|obj|so|exe|Z|elc|ln)|core)$"
)


def apply_cvsignore_filter(files):
    x = 0
    while x < len(files):
        if ignore_list.match(files[x].split("/")[-1]):
            files.pop(x)
        else:
            x += 1
    return files


def getentries(mydir, recursive=0):
    """Scans the given directory and returns a datadict of all the entries in
    the directory separated as a dirs dict and a files dict.
    """
    myfn = mydir + "/CVS/Entries"
    # entries=[dirs, files]
    entries = {"dirs": {}, "files": {}}
    if not os.path.exists(mydir):
        return entries
    try:
        myfile = open(
            _unicode_encode(myfn, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["content"],
            errors="strict",
        )
        mylines = myfile.readlines()
        myfile.close()
    except SystemExit as e:
        raise
    except:
        mylines = []

    for line in mylines:
        if line and line[-1] == "\n":
            line = line[:-1]
        if not line:
            continue
        if line == "D":  # End of entries file
            break
        mysplit = line.split("/")
        if len(mysplit) != 6:
            print("Confused:", mysplit)
            continue
        if mysplit[0] == "D":
            entries["dirs"][mysplit[1]] = {"dirs": {}, "files": {}, "status": []}
            entries["dirs"][mysplit[1]]["status"] = ["cvs"]
            if os.path.isdir(mydir + "/" + mysplit[1]):
                entries["dirs"][mysplit[1]]["status"] += ["exists"]
                entries["dirs"][mysplit[1]]["flags"] = mysplit[2:]
                if recursive:
                    rentries = getentries(mydir + "/" + mysplit[1], recursive)
                    entries["dirs"][mysplit[1]]["dirs"] = rentries["dirs"]
                    entries["dirs"][mysplit[1]]["files"] = rentries["files"]
        else:
            # [D]/Name/revision/Date/Flags/Tags
            entries["files"][mysplit[1]] = {}
            entries["files"][mysplit[1]]["revision"] = mysplit[2]
            entries["files"][mysplit[1]]["date"] = mysplit[3]
            entries["files"][mysplit[1]]["flags"] = mysplit[4]
            entries["files"][mysplit[1]]["tags"] = mysplit[5]
            entries["files"][mysplit[1]]["status"] = ["cvs"]
            if entries["files"][mysplit[1]]["revision"][0] == "-":
                entries["files"][mysplit[1]]["status"] += ["removed"]

    for file in os.listdir(mydir):
        if file == "CVS":
            continue
        if os.path.isdir(mydir + "/" + file):
            if file not in entries["dirs"]:
                if ignore_list.match(file) is not None:
                    continue
                entries["dirs"][file] = {"dirs": {}, "files": {}}
                # It's normal for a directory to be unlisted in Entries
                # when checked out without -P (see bug #257660).
                rentries = getentries(mydir + "/" + file, recursive)
                entries["dirs"][file]["dirs"] = rentries["dirs"]
                entries["dirs"][file]["files"] = rentries["files"]
            if "status" in entries["dirs"][file]:
                if "exists" not in entries["dirs"][file]["status"]:
                    entries["dirs"][file]["status"] += ["exists"]
            else:
                entries["dirs"][file]["status"] = ["exists"]
        elif os.path.isfile(mydir + "/" + file):
            if file not in entries["files"]:
                if ignore_list.match(file) is not None:
                    continue
                entries["files"][file] = {
                    "revision": "",
                    "date": "",
                    "flags": "",
                    "tags": "",
                }
            if "status" in entries["files"][file]:
                if "exists" not in entries["files"][file]["status"]:
                    entries["files"][file]["status"] += ["exists"]
            else:
                entries["files"][file]["status"] = ["exists"]
            try:
                mystat = os.stat(mydir + "/" + file)
                mytime = time.asctime(time.gmtime(mystat[stat.ST_MTIME]))
                if "status" not in entries["files"][file]:
                    entries["files"][file]["status"] = []
                if mytime == entries["files"][file]["date"]:
                    entries["files"][file]["status"] += ["current"]
            except SystemExit as e:
                raise
            except Exception as e:
                print("failed to stat", file)
                print(e)
                return

        elif ignore_list.match(file) is not None:
            pass
        else:
            print()
            print("File of unknown type:", mydir + "/" + file)
            print()

    return entries



# SOCKSv5 proxy manager for network-sandbox
# Copyright 2015-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import signal
import socket

import portage.data
from portage import _python_interpreter
from portage.data import portage_gid, portage_uid, userpriv_groups
from portage.process import atexit_register, spawn
from portage.util.futures import asyncio


class ProxyManager:
    """
    A class to start and control a single running SOCKSv5 server process
    for Portage.
    """

    def __init__(self):
        self.socket_path = None
        self._pids = []

    def start(self, settings):
        """
        Start the SOCKSv5 server.

        @param settings: Portage settings instance (used to determine
        paths)
        @type settings: portage.config
        """

        tmpdir = os.path.join(settings["PORTAGE_TMPDIR"], "portage")
        ensure_dirs_kwargs = {}
        if portage.secpass >= 1:
            ensure_dirs_kwargs["gid"] = portage_gid
            ensure_dirs_kwargs["mode"] = 0o70
            ensure_dirs_kwargs["mask"] = 0
        portage.util.ensure_dirs(tmpdir, **ensure_dirs_kwargs)

        self.socket_path = os.path.join(
            tmpdir, ".portage.%d.net.sock" % portage.getpid()
        )
        server_bin = os.path.join(settings["PORTAGE_BIN_PATH"], "socks5-server.py")
        spawn_kwargs = {}
        # The portage_uid check solves EPERM failures in Travis CI.
        if portage.data.secpass > 1 and os.geteuid() != portage_uid:
            spawn_kwargs.update(
                uid=portage_uid, gid=portage_gid, groups=userpriv_groups, umask=0o077
            )
        self._pids = spawn(
            [_python_interpreter, server_bin, self.socket_path],
            returnpid=True,
            **spawn_kwargs,
        )

    def stop(self):
        """
        Stop the SOCKSv5 server.
        """
        for p in self._pids:
            os.kill(p, signal.SIGINT)
            os.waitpid(p, 0)

        self.socket_path = None
        self._pids = []

    def is_running(self):
        """
        Check whether the SOCKSv5 server is running.

        @return: True if the server is running, False otherwise
        """
        return self.socket_path is not None

    async def ready(self):
        """
        Wait for the proxy socket to become ready. This method is a coroutine.
        """

        while True:
            try:
                wait_retval = os.waitpid(self._pids[0], os.WNOHANG)
            except OSError as e:
                if e.errno == errno.EINTR:
                    continue
                raise

            if wait_retval is not None and wait_retval != (0, 0):
                raise OSError(3, "No such process")

            try:
                s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                s.connect(self.socket_path)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                await asyncio.sleep(0.2)
            else:
                break
            finally:
                s.close()


proxy = ProxyManager()


def get_socks5_proxy(settings):
    """
    Get UNIX socket path for a SOCKSv5 proxy. A new proxy is started if
    one isn't running yet, and an atexit event is added to stop the proxy
    on exit.

    @param settings: Portage settings instance (used to determine paths)
    @type settings: portage.config
    @return: (string) UNIX socket path
    """

    if not proxy.is_running():
        proxy.start(settings)
        atexit_register(proxy.stop)

    return proxy.socket_path



# Copyright 2011-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import difflib

from portage.versions import catsplit


def similar_name_search(dbs, atom):
    cp_lower = atom.cp.lower()
    cat, pkg = catsplit(cp_lower)
    if cat == "null":
        cat = None

    all_cp = set()
    for db in dbs:
        all_cp.update(db.cp_all())

    # discard dir containing no ebuilds
    all_cp.discard(atom.cp)

    orig_cp_map = {}
    for cp_orig in all_cp:
        orig_cp_map.setdefault(cp_orig.lower(), []).append(cp_orig)
    all_cp = set(orig_cp_map)

    if cat:
        matches = difflib.get_close_matches(cp_lower, all_cp)
    else:
        pkg_to_cp = {}
        for other_cp in list(all_cp):
            other_pkg = catsplit(other_cp)[1]
            if other_pkg == pkg:
                # Check for non-identical package that
                # differs only by upper/lower case.
                identical = True
                for cp_orig in orig_cp_map[other_cp]:
                    if catsplit(cp_orig)[1] != catsplit(atom.cp)[1]:
                        identical = False
                        break
                if identical:
                    # discard dir containing no ebuilds
                    all_cp.discard(other_cp)
                    continue
            pkg_to_cp.setdefault(other_pkg, set()).add(other_cp)

        pkg_matches = difflib.get_close_matches(pkg, pkg_to_cp)
        matches = []
        for pkg_match in pkg_matches:
            matches.extend(pkg_to_cp[pkg_match])

    matches_orig_case = []
    for cp in matches:
        matches_orig_case.extend(orig_cp_map[cp])

    return matches_orig_case



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import sys
import threading

import portage.const
from portage.util import writemsg


def set_trace(on=True):
    if on:
        t = trace_handler()
        threading.settrace(t.event_handler)
        sys.settrace(t.event_handler)
    else:
        sys.settrace(None)
        threading.settrace(None)


class trace_handler:
    def __init__(self):
        python_system_paths = []
        for x in sys.path:
            if os.path.basename(x) == "python%s.%s" % sys.version_info[:2]:
                python_system_paths.append(x)

        self.ignore_prefixes = []
        for x in python_system_paths:
            self.ignore_prefixes.append(x + os.sep)

        self.trim_filename = prefix_trimmer(
            os.path.join(portage.const.PORTAGE_BASE_PATH, "lib") + os.sep
        ).trim
        self.show_local_lines = False
        self.max_repr_length = 200

    def event_handler(self, *args):
        frame, event, _arg = args
        if "line" == event:
            if self.show_local_lines:
                self.trace_line(*args)
        else:
            if not self.ignore_filename(frame.f_code.co_filename):
                self.trace_event(*args)
                return self.event_handler

    def trace_event(self, frame, event, arg):
        writemsg(
            "%s line=%d name=%s event=%s %slocals=%s\n"
            % (
                self.trim_filename(frame.f_code.co_filename),
                frame.f_lineno,
                frame.f_code.co_name,
                event,
                self.arg_repr(frame, event, arg),
                self.locals_repr(frame, event, arg),
            )
        )

    def arg_repr(self, _frame, event, arg):
        my_repr = None
        if "return" == event:
            my_repr = repr(arg)
            if len(my_repr) > self.max_repr_length:
                my_repr = "'omitted'"
            return f"value={my_repr} "
        if "exception" == event:
            my_repr = repr(arg[1])
            if len(my_repr) > self.max_repr_length:
                my_repr = "'omitted'"
            return f"type={arg[0]} value={my_repr} "

        return ""

    def trace_line(self, frame, _event, _arg):
        writemsg(
            "%s line=%d\n"
            % (self.trim_filename(frame.f_code.co_filename), frame.f_lineno)
        )

    def ignore_filename(self, filename):
        if filename:
            for x in self.ignore_prefixes:
                if filename.startswith(x):
                    return True
        return False

    def locals_repr(self, frame, _event, _arg):
        """Create a representation of the locals dict that is suitable for
        tracing output."""

        my_locals = frame.f_locals.copy()

        # prevent unsafe  __repr__ call on self when __init__ is called
        # (method calls aren't safe until after __init__  has completed).
        if frame.f_code.co_name == "__init__" and "self" in my_locals:
            my_locals["self"] = "omitted"

        # We omit items that will lead to unreasonable bloat of the trace
        # output (and resulting log file).
        for k, v in my_locals.items():
            my_repr = repr(v)
            if len(my_repr) > self.max_repr_length:
                my_locals[k] = "omitted"
        return my_locals


class prefix_trimmer:
    def __init__(self, prefix):
        self.prefix = prefix
        self.cut_index = len(prefix)
        self.previous = None
        self.previous_trimmed = None

    def trim(self, s):
        """Remove a prefix from the string and return the result.
        The previous result is automatically cached."""
        if s == self.previous:
            return self.previous_trimmed

        if s.startswith(self.prefix):
            self.previous_trimmed = s[self.cut_index :]
        else:
            self.previous_trimmed = s
        return self.previous_trimmed



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OnlydepsTestCase(TestCase):
    def testOnlydeps(self):
        ebuilds = {
            "app-misc/A-1": {
                "EAPI": "5",
                "SLOT": "1",
                "DEPEND": "|| ( app-misc/B app-misc/A:1 )",
            },
            "app-misc/A-2": {
                "EAPI": "5",
                "SLOT": "2",
            },
            "app-misc/B-0": {
                "EAPI": "5",
            },
        }

        installed = {
            "app-misc/A-2": {
                "EAPI": "5",
                "SLOT": "2",
            }
        }

        test_cases = (
            # bug 524916 - direct circular dep should not pull
            # in an onlydeps node when possible
            ResolverPlaygroundTestCase(
                ["app-misc/A:1"],
                success=True,
                options={"--onlydeps": True},
                mergelist=["app-misc/B-0"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2018-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import signal

from _emerge.AsynchronousTask import AsynchronousTask
from portage.util.futures import asyncio


class AsyncTaskFuture(AsynchronousTask):
    """
    Wraps a Future in an AsynchronousTask, which is useful for
    scheduling with TaskScheduler.
    """

    __slots__ = ("future",)

    def _start(self):
        self.future = asyncio.ensure_future(self.future, self.scheduler)
        self.future.add_done_callback(self._done_callback)

    def isAlive(self):
        """
        Returns True if self.future is an asyncio.Future that is not done.
        """
        return isinstance(self.future, asyncio.Future) and not self.future.done()

    def _cancel(self):
        if not self.future.done():
            self.future.cancel()

    def _done_callback(self, future):
        if future.cancelled():
            self.cancelled = True
            self.returncode = -signal.SIGINT
        elif future.exception() is None:
            self.returncode = os.EX_OK
        else:
            self.returncode = 1
        self._async_wait()



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.process import find_binary
from portage.util import shlex_split


def validate_cmd_var(v):
    """
    Validate an environment variable value to see if it
    contains an executable command as the first token.
    returns (valid, token_list) where 'valid' is boolean and 'token_list'
    is the (possibly empty) list of tokens split by shlex.
    """
    invalid = False
    v_split = shlex_split(v)
    if not v_split:
        invalid = True
    elif os.path.isabs(v_split[0]):
        invalid = not os.access(v_split[0], os.EX_OK)
    elif find_binary(v_split[0]) is None:
        invalid = True
    return (not invalid, v_split)



# Copyright 2011-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class CompleteGraphTestCase(TestCase):
    def testCompleteGraphUseChange(self):
        """
        Prevent reverse dependency breakage triggered by USE changes.
        """

        ebuilds = {
            "dev-libs/libxml2-2.8.0": {
                "EAPI": "2",
                "IUSE": "+icu",
                "SLOT": "2",
            },
            "x11-libs/qt-webkit-4.8.2": {
                "EAPI": "2",
                "IUSE": "icu",
                "RDEPEND": "dev-libs/libxml2:2[!icu?]",
            },
        }

        installed = {
            "dev-libs/libxml2-2.8.0": {
                "EAPI": "2",
                "IUSE": "+icu",
                "USE": "",
                "SLOT": "2",
            },
            "x11-libs/qt-webkit-4.8.2": {
                "EAPI": "2",
                "IUSE": "icu",
                "RDEPEND": "dev-libs/libxml2:2[-icu]",
                "USE": "",
            },
        }

        world = ["x11-libs/qt-webkit"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/libxml2"],
                options={"--complete-graph-if-new-use": "y"},
                mergelist=["dev-libs/libxml2-2.8.0"],
                slot_collision_solutions=[{"dev-libs/libxml2-2.8.0": {"icu": False}}],
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/libxml2"],
                options={"--complete-graph-if-new-use": "n"},
                mergelist=["dev-libs/libxml2-2.8.0"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/libxml2"],
                options={"--ignore-world": True},
                mergelist=["dev-libs/libxml2-2.8.0"],
                success=True,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testCompleteGraphVersionChange(self):
        """
        Prevent reverse dependency breakage triggered by version changes.
        """

        ebuilds = {
            "sys-libs/x-0.1": {},
            "sys-libs/x-1": {},
            "sys-libs/x-2": {},
            "sys-apps/a-1": {"RDEPEND": ">=sys-libs/x-1 <sys-libs/x-2"},
        }

        installed = {
            "sys-libs/x-1": {},
            "sys-apps/a-1": {"RDEPEND": ">=sys-libs/x-1 <sys-libs/x-2"},
        }

        world = ["sys-apps/a"]

        test_cases = (
            ResolverPlaygroundTestCase(
                [">=sys-libs/x-2"],
                options={
                    "--complete-graph-if-new-ver": "n",
                    "--rebuild-if-new-slot": "n",
                },
                mergelist=["sys-libs/x-2"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                [">=sys-libs/x-2"],
                options={"--ignore-world": True},
                mergelist=["sys-libs/x-2"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                [">=sys-libs/x-2"],
                options={"--complete-graph-if-new-ver": "y"},
                mergelist=["sys-libs/x-2"],
                slot_collision_solutions=[],
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["<sys-libs/x-1"],
                options={
                    "--complete-graph-if-new-ver": "n",
                    "--rebuild-if-new-slot": "n",
                },
                mergelist=["sys-libs/x-0.1"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["<sys-libs/x-1"],
                options={"--ignore-world": True},
                mergelist=["sys-libs/x-0.1"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["<sys-libs/x-1"],
                options={"--complete-graph-if-new-ver": "y"},
                mergelist=["sys-libs/x-0.1"],
                slot_collision_solutions=[],
                success=False,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()
